from typing import List, Tuple, Any, Optional, Dict

import numpy as np

from nebullvm.config import ONNX_PROVIDERS
from nebullvm.optional_modules.onnx import onnx
from nebullvm.optional_modules.onnxruntime import onnxruntime as ort
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.base import (
    InputInfo,
    DataType,
    DeepLearningFramework,
    Device,
)


def convert_to_numpy(tensor: Any):
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.cpu().detach().numpy()
    elif isinstance(tensor, tf.Tensor) and tensor is not None:
        tensor = tensor.numpy()
    elif isinstance(tensor, int):
        tensor = np.array([tensor])
    else:
        if not isinstance(tensor, np.ndarray):
            raise TypeError(f"Unsupported data type: {type(tensor)}")
    return tensor


def convert_to_target_framework(
    tensor: np.ndarray, framework: DeepLearningFramework
) -> Any:
    if framework is DeepLearningFramework.PYTORCH:
        return torch.from_numpy(tensor)
    elif framework is DeepLearningFramework.TENSORFLOW:
        return tf.convert_to_tensor(tensor)
    else:
        return tensor


def get_input_names(onnx_model: str):
    model = onnx.load(onnx_model)
    input_all = [node.name for node in model.graph.input]
    return input_all


def get_output_names(onnx_model: str):
    model = onnx.load(onnx_model)
    output_all = [node.name for node in model.graph.output]
    return output_all


def run_onnx_model(
    onnx_model: str, input_tensors: List[np.ndarray], device: Device
) -> List[np.ndarray]:
    from nebullvm.optional_modules.onnxruntime import onnxruntime as ort

    model = ort.InferenceSession(
        onnx_model,
        providers=ONNX_PROVIDERS["cuda"]
        if device is Device.GPU
        else ONNX_PROVIDERS["cpu"],
    )
    inputs = {
        name: array
        for name, array in zip(get_input_names(onnx_model), input_tensors)
    }
    res = model.run(
        output_names=get_output_names(onnx_model), input_feed=inputs
    )
    return list(res)


def _extract_dynamic_axis(
    onnx_model: str,
    data: List[Tuple[Tuple[np.ndarray, ...], np.ndarray]],
    input_sizes: List[Tuple[int, ...]],
    batch_size: int,
    device: Device,
    max_data: int = 100,
) -> Optional[Dict]:
    from nebullvm.tools.utils import inspect_dynamic_size

    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, input_data in enumerate(data):
        input_tensors = input_data[0]
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, batch_size, dynamic_axis["inputs"]
        )
        outputs = tuple(
            run_onnx_model(onnx_model, list(input_tensors), device)
        )
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape[1:]) for output in outputs]
        inspect_dynamic_size(
            outputs, output_sizes, batch_size, dynamic_axis["outputs"]
        )
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def extract_info_from_np_data(
    onnx_model: str,
    data: List[Tuple[Tuple[np.ndarray, ...], np.ndarray]],
    batch_size: int,
    input_sizes: List[Tuple[int, ...]],
    input_types: List[str],
    dynamic_axis: Dict,
    device: Device,
):
    from nebullvm.tools.utils import ifnone

    input_row = data[0][0]
    batch_size = ifnone(batch_size, int(input_row[0].shape[0]))
    input_sizes = ifnone(input_sizes, [tuple(x.shape[1:]) for x in input_row])
    input_types = ifnone(
        input_types,
        [
            "int32"
            if x.dtype is np.int32
            else "int64"
            if x.dtype is np.int64
            else "float32"
            for x in input_row
        ],
    )
    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(
            onnx_model, data, input_sizes, batch_size, device
        ),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def get_output_sizes_onnx(
    onnx_model: str, input_tensors: List[np.ndarray], device
) -> List[Tuple[int, ...]]:
    res = run_onnx_model(onnx_model, input_tensors, device)
    sizes = [tuple(output.shape[1:]) for output in res]
    return sizes


def create_model_inputs_onnx(
    batch_size: int, input_infos: List[InputInfo]
) -> List[np.ndarray]:
    input_tensors = (
        np.random.randn(batch_size, *input_info.size).astype(np.float32)
        if input_info.dtype is DataType.FLOAT32
        else np.random.randint(
            size=(batch_size, *input_info.size),
            low=input_info.min_value or 0,
            high=input_info.max_value or 100,
        )
        for input_info in input_infos
    )
    return list(input_tensors)


def onnx_is_gpu_available():
    return ort.get_device() == "GPU"



from nebullvm.optional_modules.torch import torch


VERSION = "0.7.3"
LEARNER_METADATA_FILENAME = "metadata.json"
ONNX_OPSET_VERSION = 13
NEBULLVM_DEBUG_FILE = "nebullvm_debug.json"

AUTO_TVM_TUNING_OPTION = {
    "tuner": "xgb",
    "trials": 10,
    "early_stopping": 100,
}
# TODO: remove the min_repeat_ms key
AUTO_TVM_PARAMS = {
    "number": 10,
    "repeat": 1,
    "min_repeat_ms": 0,  # since we're tuning on a CPU, can be set to 0
    "timeout": 10,  # in seconds
}

NVIDIA_FILENAMES = {
    "engine": "tensor_rt.engine",
    "metadata": LEARNER_METADATA_FILENAME,
}

TVM_FILENAMES = {"engine": "compiled_lib.so"}

ONNX_FILENAMES = {"model_name": "model.onnx"}
ONNX_PROVIDERS = {
    "cuda": [
        "CUDAExecutionProvider",
        "CPUExecutionProvider",
    ],
    "cpu": [
        "CPUExecutionProvider",
    ],
}

OPENVINO_FILENAMES = {
    "metadata": LEARNER_METADATA_FILENAME,
    "description_file": "description.xml",
    "weights": "weights.bin",
}

TENSORFLOW_BACKEND_FILENAMES = {
    "tflite_model": "tf_model.tflite",
    "tf_model": "tf_model.h5",
}

TORCH_TENSORRT_PRECISIONS = {
    "torch.float32": {torch.float},
    "torch.float16": {torch.float, torch.half},
    "torch.int8": {torch.float, torch.half, torch.int8},
}

MIN_DIM_INPUT_DATA = 100
QUANTIZATION_DATA_NUM = 300
CONSTRAINED_METRIC_DROP_THS = 1e-2
TRAIN_TEST_SPLIT_RATIO = 0.8

COMPILER_LIST = [
    "deepsparse",
    "tensor_rt",
    "torchscript",
    "onnxruntime",
    "tflite",
    "xla",
    "tvm",
    "openvino",
    "bladedisc",
    "intel_neural_compressor",
]

COMPRESSOR_LIST = [
    "sparseml",
    "intel_pruning",
]

ONNX_MODULES = ["openvino", "tensor_rt"]

TORCH_MODULES = [
    "deepsparse",
    "intel_neural_compressor",
    "tensor_rt",
    "torch_tensor_rt",
]

TENSORFLOW_MODULES = []
HUGGING_FACE_MODULES = []

LIBRARIES_GPU = ["tensor_rt", "torch_tensor_rt"]

MIN_NUMBER = 1e-4
DEFAULT_METRIC_DROP_THS = 1e-3
ACTIVATION_METRIC_DROP_THS = 1e-2



from typing import List, Sequence, Any

from nebullvm.config import QUANTIZATION_DATA_NUM
from nebullvm.optional_modules.tvm import (
    relay,
    ToMixedPrecision,
)
from nebullvm.tools.base import QuantizationType
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    HalfPrecisionTransformation,
)


class TVMCalibrator(DataManager):
    def __init__(self, data_reader: Sequence, input_names: List[str]):
        super(TVMCalibrator, self).__init__(data_reader=data_reader)
        self._input_names = input_names

    def __getitem__(self, item: int):
        tuple_ = self._data_reader[item]
        return {name: data for name, data in zip(self._input_names, tuple_)}


def quantize_apache_tvm(
    model: Any,
    quantization_type: QuantizationType,
    input_tfms: MultiStageTransformation,
    input_data: DataManager,
    params: Any,
):
    if quantization_type is not None:
        if quantization_type is QuantizationType.HALF:
            quantized_model = ToMixedPrecision(mixed_precision_type="float16")(
                model
            )
            input_tfms.append(HalfPrecisionTransformation())
        else:
            if quantization_type is QuantizationType.DYNAMIC:
                inputs = None
            elif quantization_type is QuantizationType.STATIC:
                inputs = input_data.get_split("train").get_numpy_list(
                    QUANTIZATION_DATA_NUM
                )
                input_names = [f"input_{n}" for n in range(len(inputs[0]))]
                inputs = TVMCalibrator(inputs, input_names)
            else:
                return

            if inputs is not None:
                with relay.quantize.qconfig(
                    calibrate_mode="kl_divergence", weight_scale="max"
                ):
                    quantized_model = relay.quantize.quantize(
                        model, params, dataset=inputs
                    )
            else:
                with relay.quantize.qconfig(
                    calibrate_mode="global_scale", global_scale=8.0
                ):
                    quantized_model = relay.quantize.quantize(model, params)

        return quantized_model



from pathlib import Path
from typing import Union, List, Tuple

import numpy as np

from nebullvm.config import QUANTIZATION_DATA_NUM
from nebullvm.operations.optimizations.compilers.base import Compiler

from nebullvm.operations.optimizations.compilers.quantizations.onnx import (
    quantize_onnx,
)
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.tools.base import QuantizationType
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class ONNXCompiler(Compiler):
    supported_ops = {
        "cpu": [
            None,
            QuantizationType.STATIC,
            QuantizationType.HALF,
            QuantizationType.DYNAMIC,
        ],
        "gpu": [
            None,
            QuantizationType.STATIC,
            QuantizationType.HALF,
            QuantizationType.DYNAMIC,
        ],
    }

    def execute(
        self,
        model: str,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Compile the input model using ONNX Runtime Compiler.

        Args:
            model (str): The onnx model path.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)
        train_input_data = input_data.get_split("train").get_numpy_list(
            QUANTIZATION_DATA_NUM
        )

        if quantization_type is not None:
            model = self._quantize_model(
                model, train_input_data, quantization_type, input_tfms
            )

        self.compiled_model = self._compile_model(model)

    def _compile_model(self, model: Union[str, Path]):
        return model

    def _quantize_model(
        self,
        model_path: str,
        input_data: List[Tuple[np.ndarray, ...]],
        quantization_type: QuantizationType,
        input_tfms: MultiStageTransformation,
    ):
        return quantize_onnx(
            model_path, input_data, quantization_type, self.device, input_tfms
        )



import time
from abc import abstractmethod, ABC
from typing import Any, Dict, Type

import numpy as np
from loguru import logger
from tqdm import tqdm

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch, DataLoader
from nebullvm.tools.base import DeepLearningFramework, ModelParams
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import create_model_inputs_onnx
from nebullvm.tools.pytorch import create_model_inputs_torch
from nebullvm.tools.tf import create_model_inputs_tf
from nebullvm.tools.utils import (
    check_input_data,
    extract_info_from_data,
    is_data_subscriptable,
)


def _get_dl_framework(model: Any):
    if isinstance(model, torch.nn.Module) or str(model).startswith("Pytorch"):
        return DeepLearningFramework.PYTORCH
    elif (isinstance(model, tf.Module) and model is not None) or str(
        model
    ).startswith("Tensorflow"):
        return DeepLearningFramework.TENSORFLOW
    elif isinstance(model, str) or str(model).startswith("Numpy"):
        return DeepLearningFramework.NUMPY
    else:
        raise TypeError(f"Model type {type(model)} not supported.")


def _create_model_inputs(
    dl_framework: DeepLearningFramework, model_params: ModelParams
):
    if dl_framework == DeepLearningFramework.PYTORCH:
        input_data = create_model_inputs_torch(
            model_params.batch_size, model_params.input_infos
        )
    elif dl_framework == DeepLearningFramework.TENSORFLOW:
        input_data = create_model_inputs_tf(
            model_params.batch_size, model_params.input_infos
        )
    elif dl_framework == DeepLearningFramework.NUMPY:
        input_data = create_model_inputs_onnx(
            model_params.batch_size, model_params.input_infos
        )
    else:
        raise TypeError(f"Unknown framework {dl_framework}")

    return input_data


class BaseBenchmark(ABC):
    def __init__(self, model, input_tensors, n_warmup=50, n_runs=1000):
        self.model = model
        self.input_tensors = input_tensors
        self.n_warmup = n_warmup
        self.n_runs = n_runs

    @abstractmethod
    def benchmark(self):
        raise NotImplementedError


class PytorchBenchmark(BaseBenchmark):
    def benchmark(self):
        has_cuda = torch.cuda.is_available()
        device = torch.device("cuda" if has_cuda else "cpu")
        input_tensors = [
            [tensor.to(device) for tensor in tensors]
            for tensors in self.input_tensors
        ]
        batch_size = input_tensors[0][0].shape[0]

        if isinstance(self.model, torch.nn.Module):
            self.model.to(device).eval()

        if isinstance(self.model, torch.nn.Module):
            self.model.to(device).eval()

        with torch.no_grad():
            for i in tqdm(
                range(self.n_warmup),
                desc=f"Performing warm up on {self.n_warmup} iterations",
            ):
                self.model(
                    *input_tensors[i % min(self.n_warmup, len(input_tensors))]
                )
        if has_cuda:
            torch.cuda.synchronize()
        timings = []
        with torch.no_grad():
            for i in tqdm(
                range(1, self.n_runs + 1),
                desc=f"Performing benchmark on {self.n_runs} iterations",
            ):
                start_time = time.time()
                self.model(
                    *input_tensors[i % min(self.n_runs, len(input_tensors))]
                )
                if has_cuda:
                    torch.cuda.synchronize()
                end_time = time.time()
                timings.append(end_time - start_time)

        print(f"Batch size: {batch_size}")

        throughput = batch_size / np.mean(timings)
        latency = np.mean(timings) / batch_size

        print("Average Throughput: %.2f data/second" % throughput)
        print("Average Latency: %.4f seconds/data" % latency)

        return throughput, latency


class TensorflowBenchmark(BaseBenchmark):
    def benchmark(self):
        # TODO: implement benchmark for tensorflow
        raise NotImplementedError


class NumpyBenchmark(BaseBenchmark):
    def benchmark(self):
        # TODO: implement benchmark for numpy
        raise NotImplementedError


def benchmark(
    model, input_data, device=None, random=False, n_warmup=50, n_runs=1000
):
    """Performs a Benchmark on the input model regardless of the framework it
    was used for implementing it.
    Args:
        model (Any): The input model.
        input_data (Iterable or Sequence): Input data to be used for
            optimizing the model. PyTorch, TensorFlow
            and Onnx respectively accept input tensor in `torch.Tensor`,
            `tf.Tensor` and `np.ndarray` formats. Note that the each input
            sample must be a tuple containing a tuple as first element, the
            `inputs`, and the `label` as second element. The `inputs` needs to
            be passed as tuple even if a single input is needed by the model
            (in this case the `inputs` tuple will contain just an element).
            HuggingFace models can take as data samples both dictionaries or
            strings. Strings will then be converted in data samples using the
            HuggingFace tokenizer which must be given as input when just a
            list of string is provided as input_data (tokenizers can be passed
            as extra arguments of this function using the keyword `tokenizer`).
        device (str): Device to be used for running the benchmark. If None,
            CPU will be used. Default: None.
        random (bool, optional): If set to true, the data used to benchmark the
            model will be computed randomly given the info extracted from the
            provided input_data.
        n_warmup (int, optional): Number of warmup iterations.
        n_runs (int, optional): Number of iterations performed to benchmark
            the model.
    """
    if device is None:
        logger.warning("No device specified, using CPU")
        device = "cpu"

    dl_framework = _get_dl_framework(model)

    if isinstance(input_data, (DataLoader, tf.data.Dataset)):
        try:
            input_data = DataManager.from_dataloader(input_data)
        except Exception:
            raise ValueError(
                "The provided dataloader does not match the expected "
                "format.\n"
                "Speedster supports dataloaders that return tuples in "
                "the\n"
                "following formats: \n"
                "Single input: (input,  label)\n"
                "Multiple inputs: ((input1, input2, ...),  label) or "
                "(input1, input2, ...,  label)\n"
                "Inputs and labels should be either tensors or numpy "
                "arrays,\n"
                "depending on the framework used.\n"
            )

    if not isinstance(input_data, DataManager):
        if check_input_data(input_data):
            if is_data_subscriptable(input_data):
                input_data = DataManager(input_data)
            else:
                input_data = DataManager.from_iterable(input_data)
        else:
            raise ValueError(
                "The provided data does not match the expected "
                "format.\n"
                "Speedster supports data in the following formats: \n"
                "- PyTorch DataLoader\n"
                "- TensorFlow Dataset\n"
                "- List of tuples: [((input_0, ... ), label), ...] \n"
                "Inputs and labels should be either tensors or numpy "
                "arrays,\n"
                "depending on the framework used.\n"
            )

    if random:
        model_params = extract_info_from_data(
            model, input_data, dl_framework, None, device
        )
        input_data = _create_model_inputs(dl_framework, model_params)
    else:
        input_data = input_data.get_list()

    BENCHMARK_FUNCTIONS[dl_framework](
        model=model,
        input_tensors=input_data,
        n_warmup=n_warmup,
        n_runs=n_runs,
    ).benchmark()


BENCHMARK_FUNCTIONS: Dict[DeepLearningFramework, Type[BaseBenchmark]] = {
    DeepLearningFramework.PYTORCH: PytorchBenchmark,
    DeepLearningFramework.TENSORFLOW: TensorflowBenchmark,
    DeepLearningFramework.NUMPY: NumpyBenchmark,
}



from pathlib import Path
from typing import Union

from nebullvm.operations.conversions.converters import (
    PytorchConverter,
)
from nebullvm.operations.optimizations.compilers.base import Compiler
from nebullvm.optional_modules.torch import (
    Module,
    GraphModule,
)
from nebullvm.tools.base import (
    ModelParams,
    QuantizationType,
)
from nebullvm.tools.data import DataManager


class DeepSparseCompiler(Compiler):
    supported_ops = {
        "cpu": [None],
        "gpu": [],
    }

    def __init__(self):
        super().__init__()
        self.conversion_op = PytorchConverter()

    def execute(
        self,
        model: Module,
        onnx_output_path: str,
        model_params: ModelParams,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Compile the input model using DeepSparse Compiler.

        Args:
            model (torch.nn.Module): The pytorch model.
            onnx_output_path (str): Path where the converted ONNX model will be
                stored.
            model_params (ModelParams): The model parameters.
            quantization_type (QuantizationType): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        self.compiled_model = self._compile_model(
            model, onnx_output_path, input_data, model_params
        )

    def _compile_model(
        self,
        model: Union[Module, GraphModule],
        onnx_output_path: str,
        input_data: DataManager,
        model_params: ModelParams,
    ) -> str:
        self.conversion_op.model_name = "model_pruned"
        onnx_pruned_path = Path(onnx_output_path)
        self.conversion_op.to(self.device).set_state(
            model, input_data
        ).execute(onnx_pruned_path, model_params)
        onnx_pruned_path = str(onnx_pruned_path / "model_pruned.onnx")
        return onnx_pruned_path

    @staticmethod
    def _quantize_model(**kwargs):
        raise NotImplementedError()



import time
from typing import Tuple, List, Union, Any

import numpy as np
from loguru import logger

from nebullvm.config import ONNX_PROVIDERS
from nebullvm.operations.inference_learners.base import BaseInferenceLearner
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch, Module
from nebullvm.tools.base import Device
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import (
    convert_to_numpy,
    get_input_names,
    get_output_names,
)


def compute_torch_latency(
    xs: List[Tuple[torch.Tensor]],
    model: Module,
    device: Device,
    steps: int = 100,
    warmup_steps: int = 10,
) -> Tuple[float, List[float]]:
    """Compute the latency associated with the torch model.

    Args:
        xs (List[Tuple[torch.Tensor]]): List of tuples containing the
            input tensors (a single batch for the model).
        model (Module): Torch model.
        device (Device): Device where computing the latency.
        steps (int, optional): Number of input data to be used to compute the
            latency of the model. It must be a number <= len(xs). Default: 100.
        warmup_steps (int, optional): Number of input data to be used to warm
            up the model. It must be a number <= len(xs). Default: 10.

    Returns:
        Float: Average latency.
        List[Float]: List of latencies obtained.
    """
    device = "cuda" if device is Device.GPU else "cpu"
    xs = [tuple(t.to(device) for t in tensors) for tensors in xs]
    model = model.to(device).eval()
    latencies = []
    with torch.no_grad():
        for i in range(warmup_steps):
            _ = model.forward(*xs[i])
        for i in range(steps):
            starting_time = time.time()
            _ = model.forward(*xs[i])
            latencies.append(time.time() - starting_time)
        latency = np.mean(latencies)
    return latency, latencies


def compute_tf_latency(
    xs: List[Tuple[tf.Tensor]],
    model: Union[tf.Module, tf.keras.Model],
    device: Device,
    steps: int = 100,
    warmup_steps: int = 10,
) -> Tuple[float, List[float]]:
    """Compute the latency associated with the tensorflow model.

    Args:
        xs (List[Tuple[tf.Tensor]]): List of tuples containing the
            input tensors (a single batch for the model).
        model (Module or keras.Model): TF model.
        device (Device): Device where computing the latency.
        steps (int, optional): Number of input data to be used to compute the
            latency of the model. It must be a number <= len(xs). Default: 100.
        warmup_steps (int, optional): Number of input data to be used to warm
            up the model. It must be a number <= len(xs). Default: 10.

    Returns:
        Float: Average latency.
        List[Float]: List of latencies obtained.
    """
    latencies = []
    with tf.device(device.value):
        for i in range(warmup_steps):
            _ = model(xs[i])
        for i in range(steps):
            starting_time = time.time()
            _ = model(xs[i])
            latencies.append(time.time() - starting_time)
        latency = np.mean(latencies)
        return latency, latencies


def compute_onnx_latency(
    xs: List[Tuple[np.array]],
    model: str,
    device: Device,
    steps: int = 100,
    warmup_steps: int = 10,
) -> Tuple[float, List[float]]:
    """Compute the latency associated with the ONNX model.

    Args:
        xs (List[Tuple[np.array]]): List of tuples containing the
            inputs (a single batch for the model).
        model (str): ONNX model path.
        device (Device): Device where computing the latency.
        steps (int, optional): Number of input data to be used to compute the
            latency of the model. It must be a number <= len(xs). Default: 100.
        warmup_steps (int, optional): Number of input data to be used to warm
            up the model. It must be a number <= len(xs). Default: 10.

    Returns:
        Float: Average latency.
        List[Float]: List of latencies obtained.
    """
    from nebullvm.optional_modules.onnxruntime import onnxruntime as ort

    input_names = get_input_names(model)
    output_names = get_output_names(model)

    model = ort.InferenceSession(
        model,
        providers=ONNX_PROVIDERS["cuda"]
        if device is Device.GPU
        else ONNX_PROVIDERS["cpu"],
    )

    latencies = []
    for i in range(warmup_steps):
        inputs = {name: array for name, array in zip(input_names, xs[i])}
        _ = model.run(output_names=output_names, input_feed=inputs)
    for i in range(steps):
        inputs = {name: array for name, array in zip(input_names, xs[i])}
        starting_time = time.time()
        _ = model.run(output_names=output_names, input_feed=inputs)
        latencies.append(time.time() - starting_time)
    latency = np.mean(latencies)
    return latency, latencies


def compute_optimized_running_time(
    optimized_model: BaseInferenceLearner,
    input_data: DataManager,
    steps: int = 100,
    min_steps: int = 5,
    warmup_steps: int = 10,
) -> float:
    """Compute the running time of the optimized model.

    Args:
        optimized_model (BaseInferenceLearner): Optimized model.
        input_data: (DataManager): Dataset used to compute latency.
        steps (int, optional): Number of input data to be used to
            compute the latency of the model. Default: 100.
        min_steps (int, optional): Minimum number of iterations to
            be performed. Default: 5.
        warmup_steps (int, optional): Number of input data to be used
            to warm up the model. Default: 10.

    Returns:
        Float: Average latency.
    """

    latencies = []
    last_median = None

    # Warmup
    inputs_list = input_data.get_split("test").get_list(warmup_steps)
    for model_inputs in inputs_list:
        _ = optimized_model(*model_inputs)

    # Compute latency
    inputs_list = input_data.get_split("test").get_list(steps)
    for model_inputs in inputs_list:
        starting_time = time.time()
        _ = optimized_model(*model_inputs)
        latencies.append(time.time() - starting_time)
        if len(latencies) > min_steps:
            median = np.median(latencies)
            diff = (
                np.abs(median - last_median) / last_median
                if last_median is not None
                else 1.0
            )
            if diff < 0.05:
                return median
            last_median = median
    return np.median(latencies)


def compute_relative_difference(
    tensor_1: Any,
    tensor_2: Any,
    y: Any = None,
    eps: float = 1e-5,
) -> float:
    if y is not None:
        logger.debug(
            "Received a label for the precision computation. "
            "It will be ignored."
        )

    tensor_1, tensor_2 = map(convert_to_numpy, (tensor_1, tensor_2))

    assert tensor_1.shape == tensor_2.shape, (
        "The outputs of the original and optimized models have "
        "different shapes"
    )

    diff = np.abs(tensor_1 - tensor_2) / (
        np.maximum(np.abs(tensor_1), np.abs(tensor_2)) + eps
    )
    return float(np.mean(diff))


def compute_accuracy_drop(tensor_1: Any, tensor_2: Any, y: Any) -> float:
    assert y is not None, (
        "No label found in the dataloader provided. "
        "To use accuracy metric, you must set also the labels"
    )
    tensor_1, tensor_2, y = map(convert_to_numpy, (tensor_1, tensor_2, y))
    accuracy_1 = np.mean(tensor_1.argmax(axis=-1) == y)
    accuracy_2 = np.mean(tensor_2.argmax(axis=-1) == y)
    return accuracy_1 - accuracy_2


QUANTIZATION_METRIC_MAP = {
    "accuracy": compute_accuracy_drop,
    "numeric_precision": compute_relative_difference,
}



from typing import Any, Union, Iterable, Sequence

from nebullvm.operations.base import Operation


class FetchModelFromLocal(Operation):
    def execute(self, model: Any):
        self.state["model"] = model

    def get_model(self) -> any:
        return self.state.get("model")

    def get_result(self) -> Any:
        pass


class FetchDataFromLocal(Operation):
    def execute(self, data: Union[Iterable, Sequence]):
        self.state["data"] = data

    def get_data(self) -> any:
        return self.state.get("data")

    def get_result(self) -> Any:
        pass



import json
import os
from pathlib import Path
from typing import Any

import requests

from nebullvm.config import VERSION

NEBULLVM_METADATA_PATH = Path.home() / ".nebullvm/collect.json"


class FeedbackCollector:
    def __init__(
        self, url: str, disable_telemetry_environ_var: str, app_version: str
    ):
        self._disable_telemetry_environ_var = disable_telemetry_environ_var
        self._is_active = (
            int(os.getenv(disable_telemetry_environ_var, "0")) == 0
        )
        self._url = url
        self._metadata = {
            "nebullvm_version": VERSION,
            "app_version": app_version,
        }

    def _store_ip_address(self):
        try:
            self._metadata["ip_address"] = requests.get(
                "https://api.ipify.org"
            ).text
        except Exception:
            self._metadata["ip_address"] = "Unknown"

    @property
    def is_active(self):
        return self._is_active

    def _inform_user(self):
        message = (
            f"Nebuly collects anonymous usage statistics to help improve the "
            f"product. You can opt-out by setting the environment variable "
            f"{self._disable_telemetry_environ_var}=1."
        )
        print(message)

    def store_info(self, key: str, value: Any):
        if key in self._metadata and isinstance(value, list):
            self._metadata[key] += value
        else:
            self._metadata[key] = value

    def send_feedback(self, timeout: int = 30):
        if not self.is_active:
            return {}
        self._store_ip_address()
        request_body = self._metadata
        headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
        }
        response = requests.post(
            self._url,
            data=json.dumps(request_body),
            headers=headers,
            timeout=timeout,
        )
        return response

    def get(self, key: str, default: Any = None):
        return self._metadata.get(key, default)

    def reset(self, key: str):
        self._metadata.pop(key, None)



from tempfile import TemporaryDirectory

import pytest

from nebullvm.config import CONSTRAINED_METRIC_DROP_THS
from nebullvm.operations.inference_learners.deepsparse import (
    DEEPSPARSE_INFERENCE_LEARNERS,
)
from nebullvm.operations.measures.measures import MetricDropMeasure
from nebullvm.operations.measures.utils import compute_relative_difference
from nebullvm.operations.optimizations.base import (
    COMPILER_TO_INFERENCE_LEARNER_MAP,
)
from nebullvm.operations.optimizations.compilers.deepsparse import (
    DeepSparseCompiler,
)
from nebullvm.operations.optimizations.compilers.utils import (
    deepsparse_is_available,
)
from nebullvm.operations.optimizations.tests.utils import initialize_model
from nebullvm.tools.base import DeepLearningFramework, Device, ModelCompiler

device = Device.CPU


@pytest.mark.parametrize(
    ("output_library", "dynamic"),
    [
        # (DeepLearningFramework.PYTORCH, True),
        (DeepLearningFramework.PYTORCH, False),
    ],
)
@pytest.mark.skipif(
    not deepsparse_is_available(),
    reason="Can't test deepsparse if it's not installed.",
)
def test_deepsparse(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type=None,
):
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            input_data,
            model_params,
            input_tfms,
            model_outputs,
            metric,
        ) = initialize_model(dynamic, None, output_library, device)

        compiler_op = DeepSparseCompiler()
        compiler_op.to(device).execute(
            model=model,
            onnx_output_path=tmp_dir,
            model_params=model_params,
            quantization_type=None,
            input_data=input_data,
        )

        compiled_model = compiler_op.get_result()

        build_inference_learner_op = COMPILER_TO_INFERENCE_LEARNER_MAP[
            ModelCompiler.DEEPSPARSE
        ][DeepLearningFramework.PYTORCH]()

        build_inference_learner_op.to(device).execute(
            model=compiled_model,
            model_orig=compiler_op.model_orig
            if hasattr(compiler_op, "model_orig")
            else None,
            model_params=model_params,
            input_tfms=input_tfms,
            source_dl_framework=output_library,
        )

        optimized_model = build_inference_learner_op.get_result()
        assert isinstance(
            optimized_model, DEEPSPARSE_INFERENCE_LEARNERS[output_library]
        )
        assert isinstance(optimized_model.get_size(), int)

        # Test save and load functions
        optimized_model.save(tmp_dir)
        loaded_model = DEEPSPARSE_INFERENCE_LEARNERS[output_library].load(
            tmp_dir
        )
        assert isinstance(
            loaded_model, DEEPSPARSE_INFERENCE_LEARNERS[output_library]
        )

        inputs_example = optimized_model.get_inputs_example()
        res = optimized_model(*inputs_example)
        assert res is not None

        # Test validity of the model
        test_input_data, ys = input_data.get_split("test").get_list(
            with_ys=True
        )

        validity_check_op = MetricDropMeasure()
        validity_check_op.execute(
            optimized_model,
            test_input_data,
            model_outputs,
            CONSTRAINED_METRIC_DROP_THS,
            metric_func=metric
            if quantization_type is not None
            else compute_relative_difference,
            ys=ys,
        )

        # Check validity of the optimized model
        assert validity_check_op.get_result()

        # Dynamic batch size is currently not supported from deepsparse
        # if dynamic:
        #     inputs_example = [
        #         input_[: len(input_) // 2] for input_ in inputs_example
        #     ]
        #     res = model(*inputs_example)
        #     assert res is not None



from tempfile import TemporaryDirectory

import pytest

from nebullvm.operations.inference_learners.neural_compressor import (
    NEURAL_COMPRESSOR_INFERENCE_LEARNERS,
)
from nebullvm.operations.optimizations.base import (
    COMPILER_TO_INFERENCE_LEARNER_MAP,
)
from nebullvm.operations.optimizations.compilers.intel_neural_compressor import (  # noqa: E501
    IntelNeuralCompressorCompiler,
)
from nebullvm.operations.optimizations.compilers.utils import (
    intel_neural_compressor_is_available,
)
from nebullvm.operations.optimizations.tests.utils import (
    initialize_model,
    check_model_validity,
)
from nebullvm.tools.base import (
    DeepLearningFramework,
    QuantizationType,
    Device,
    ModelCompiler,
)

device = Device.CPU


@pytest.mark.parametrize(
    ("output_library", "dynamic", "metric_drop_ths", "quantization_type"),
    [
        (DeepLearningFramework.PYTORCH, True, 2, QuantizationType.DYNAMIC),
        (DeepLearningFramework.PYTORCH, False, 2, QuantizationType.DYNAMIC),
        (DeepLearningFramework.PYTORCH, True, 2, QuantizationType.STATIC),
        (DeepLearningFramework.PYTORCH, False, 2, QuantizationType.STATIC),
    ],
)
@pytest.mark.skipif(
    not intel_neural_compressor_is_available(),
    reason="Can't test neural compressor if it's not installed.",
)
def test_neural_compressor(
    output_library: DeepLearningFramework,
    dynamic: bool,
    metric_drop_ths: float,
    quantization_type: QuantizationType,
):
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            input_data,
            model_params,
            input_tfms,
            model_outputs,
            metric,
        ) = initialize_model(dynamic, None, output_library, device)

        compiler_op = IntelNeuralCompressorCompiler()
        compiler_op.to(device).execute(
            model=model,
            input_tfms=input_tfms,
            metric_drop_ths=metric_drop_ths,
            quantization_type=quantization_type,
            input_data=input_data,
        )

        compiled_model = compiler_op.get_result()

        build_inference_learner_op = COMPILER_TO_INFERENCE_LEARNER_MAP[
            ModelCompiler.INTEL_NEURAL_COMPRESSOR
        ][DeepLearningFramework.PYTORCH]()

        build_inference_learner_op.to(device).execute(
            model=compiled_model,
            model_orig=compiler_op.model_orig
            if hasattr(compiler_op, "model_orig")
            else None,
            model_params=model_params,
            input_tfms=input_tfms,
            source_dl_framework=output_library,
        )

        optimized_model = build_inference_learner_op.get_result()

        assert isinstance(
            optimized_model,
            NEURAL_COMPRESSOR_INFERENCE_LEARNERS[output_library],
        )

        # Test save and load functions
        optimized_model.save(tmp_dir)
        loaded_model = NEURAL_COMPRESSOR_INFERENCE_LEARNERS[
            output_library
        ].load(tmp_dir)
        assert isinstance(
            loaded_model, NEURAL_COMPRESSOR_INFERENCE_LEARNERS[output_library]
        )

        assert isinstance(optimized_model.get_size(), int)

        inputs_example = optimized_model.get_inputs_example()
        res = optimized_model(*inputs_example)
        assert res is not None

        # Test validity of the model
        valid = check_model_validity(
            optimized_model,
            input_data,
            model_outputs,
            metric_drop_ths,
            quantization_type,
            metric,
        )
        assert valid

        if dynamic:
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = model(*inputs_example)
            assert res is not None



import logging
import os
import sys
import warnings
from typing import Any

from loguru import logger


levels_map = {
    0: "ERROR",
    1: "WARNING",
    2: "INFO",
    3: "DEBUG",
}


def debug_mode_enabled():
    return int(os.environ.get("DEBUG_MODE", "0")) > 0


def setup_logger():
    if not debug_mode_enabled():
        warnings.filterwarnings("ignore")

    logging_level = int(os.environ.get("NEBULLVM_LOG_LEVEL", "2"))

    logger.remove()
    logger.add(
        sys.stdout,
        colorize=True,
        format=(
            "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
            "<level>{level: <8}</level> | <level>{message}</level>"
        ),
        level=levels_map[logging_level],
    )
    logger.level("WARNING", color="<fg #d3d3d3>")


class LoggingContext(object):
    def __init__(
        self,
        logger: logging.Logger,
        disabled: bool = False,
        handler: Any = None,
        close: bool = True,
    ):
        self.logger = logger
        self.disabled = disabled
        self.handler = handler
        self.close = close

    def __enter__(self):
        self.logger.disabled = self.disabled
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et: Any, ev: Any, tb: Any):
        if self.disabled is True:
            self.logger.disabled = False
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import yaml

from nebullvm.optional_modules.neural_compressor import (
    MixedPrecision,
    Quantization,
)
from nebullvm.optional_modules.torch import DataLoader, Module, GraphModule
from nebullvm.tools.base import QuantizationType
from nebullvm.tools.data import DataManager, PytorchDataset
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    HalfPrecisionTransformation,
)


def _prepare_quantization_config(model: Any, tmp_dir: str, approach: str):
    config = {
        "model": {
            "name": model.__class__.__name__,
            "framework": "pytorch_fx",
        },
        "quantization": {"approach": approach},
        "evaluation": {"accuracy": {"metric": {"topk": 1}}},
        "tuning": {
            "accuracy_criterion": {"relative": 0.01},
        },
    }

    path_file = Path(tmp_dir) / "temp_qt.yaml"
    with open(path_file, "w") as f:
        yaml.dump(config, f)

    return path_file


def _prepare_mixed_precision_config(model: Any, tmp_dir: str):
    config = {
        "model": {
            "name": model.__class__.__name__,
            "framework": "pytorch_fx",
        },
        "mixed_precision": {"precisions": "bf16"},
        "evaluation": {"accuracy": {"metric": {"topk": 1}}},
        "tuning": {
            "accuracy_criterion": {"relative": 0.01},
        },
    }

    path_file = Path(tmp_dir) / "temp_mp.yaml"
    with open(path_file, "w") as f:
        yaml.dump(config, f)

    return path_file


def _get_dataloader(input_data: DataManager):
    bs = input_data[0][0][0].shape[0]
    ds = PytorchDataset(input_data, has_labels=True)
    dl = DataLoader(ds, bs)
    return dl


def _quantize_static(model: Module, input_data: DataManager) -> GraphModule:
    with TemporaryDirectory() as tmp_dir:
        config_file_qt = _prepare_quantization_config(
            model, tmp_dir, "post_training_static_quant"
        )
        quantizer = Quantization(str(config_file_qt))
        quantizer.model = model
        quantizer.calib_dataloader = _get_dataloader(input_data)
        quantizer.eval_dataloader = _get_dataloader(input_data)
        compressed_model = quantizer()

    return compressed_model


def _quantize_dynamic(model: Module) -> GraphModule:
    with TemporaryDirectory() as tmp_dir:
        config_file_qt = _prepare_quantization_config(
            model, tmp_dir, "post_training_dynamic_quant"
        )
        quantizer = Quantization(str(config_file_qt))
        quantizer.model = model
        compressed_model = quantizer()

    return compressed_model


def _mixed_precision(
    model: Module, input_tfms: MultiStageTransformation
) -> GraphModule:
    with TemporaryDirectory() as tmp_dir:
        config_file_qt = _prepare_mixed_precision_config(model, tmp_dir)
        converter = MixedPrecision(str(config_file_qt))
        converter.model = model
        compressed_model = converter()

    input_tfms.append(HalfPrecisionTransformation())

    return compressed_model


def quantize_neural_compressor(
    model: Module,
    quantization_type: QuantizationType,
    input_tfms: MultiStageTransformation,
    input_data: DataManager,
) -> GraphModule:
    if quantization_type is QuantizationType.STATIC:
        quantized_model = _quantize_static(model, input_data)
    elif quantization_type is QuantizationType.DYNAMIC:
        quantized_model = _quantize_dynamic(model)
    elif quantization_type is QuantizationType.HALF:
        quantized_model = _mixed_precision(model, input_tfms)
    else:
        raise ValueError(
            f"Quantization type {quantization_type} is not "
            f"supported by Intel Neural Compressor"
        )
    return quantized_model



import abc
from tempfile import TemporaryDirectory
from typing import List, Callable, Union, Tuple, Any, Dict, Type

from nebullvm.config import (
    CONSTRAINED_METRIC_DROP_THS,
    ACTIVATION_METRIC_DROP_THS,
)
from nebullvm.operations.base import Operation
from nebullvm.operations.inference_learners.base import BuildInferenceLearner
from nebullvm.operations.inference_learners.builders import (
    PytorchBuildInferenceLearner,
    DeepSparseBuildInferenceLearner,
    IntelNeuralCompressorBuildInferenceLearner,
    PyTorchTensorRTBuildInferenceLearner,
    ONNXTensorRTBuildInferenceLearner,
    PyTorchApacheTVMBuildInferenceLearner,
    ONNXApacheTVMBuildInferenceLearner,
    ONNXBuildInferenceLearner,
    OpenVINOBuildInferenceLearner,
    TFLiteBuildInferenceLearner,
    TensorflowBuildInferenceLearner,
)
from nebullvm.operations.measures.measures import MetricDropMeasure
from nebullvm.operations.measures.utils import (
    compute_relative_difference,
    compute_optimized_running_time,
)
from nebullvm.operations.optimizations.compilers.base import Compiler
from nebullvm.operations.optimizations.compilers.deepsparse import (
    DeepSparseCompiler,
)
from nebullvm.operations.optimizations.compilers.intel_neural_compressor import (  # noqa: E501
    IntelNeuralCompressorCompiler,
)
from nebullvm.operations.optimizations.compilers.onnxruntime import (
    ONNXCompiler,
)
from nebullvm.operations.optimizations.compilers.openvino import (
    OpenVINOCompiler,
)
from nebullvm.operations.optimizations.compilers.pytorch import (
    PytorchBackendCompiler,
)
from nebullvm.operations.optimizations.compilers.tensor_rt import (
    PyTorchTensorRTCompiler,
    ONNXTensorRTCompiler,
)
from nebullvm.operations.optimizations.compilers.tensorflow import (
    TFLiteBackendCompiler,
    TensorflowBackendCompiler,
)
from nebullvm.operations.optimizations.compilers.tvm import (
    PyTorchApacheTVMCompiler,
    ONNXApacheTVMCompiler,
)

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.base import (
    ModelCompiler,
    QuantizationType,
    OptimizationTime,
    ModelParams,
    DeepLearningFramework,
    ModelCompressor,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class Optimizer(Operation, abc.ABC):
    def __init__(self):
        super().__init__()
        self.optimized_models = []
        self.source_dl_framework = None
        self.pipeline_dl_framework = None
        self.compiler_ops = {}
        self.build_inference_learner_ops = {}
        self.validity_check_op = MetricDropMeasure()

    def execute(
        self,
        model: Any,
        input_data: DataManager,
        optimization_time: OptimizationTime,
        metric_drop_ths: str,
        metric: Callable,
        model_params: ModelParams,
        model_outputs: List[Tuple[Any, ...]],
        ignore_compilers: List[ModelCompiler],
        ignore_compressors: List[ModelCompressor],
        source_dl_framework: DeepLearningFramework,
    ):
        self.source_dl_framework = source_dl_framework

        # TODO: implement and select compressors from hardware

        compilers = self._select_compilers_from_hardware()
        (
            self.compiler_ops,
            self.build_inference_learner_ops,
        ) = self._load_compilers(
            ignore_compilers=ignore_compilers,
            compilers=compilers,
        )
        self._optimize(
            model=model,
            input_data=input_data,
            optimization_time=optimization_time,
            metric_drop_ths=metric_drop_ths,
            metric=metric,
            model_params=model_params,
            model_outputs=model_outputs,
            ignore_compilers=ignore_compilers,
        )

    @abc.abstractmethod
    def _select_compilers_from_hardware(self):
        raise NotImplementedError()

    def _load_compilers(
        self,
        ignore_compilers: List[ModelCompiler],
        compilers: List[ModelCompiler],
    ):
        compiler_ops = {
            compiler: COMPILER_TO_OPTIMIZER_MAP[compiler][
                self.pipeline_dl_framework
            ]()
            for compiler in compilers
            if compiler not in ignore_compilers
            and compiler in COMPILER_TO_OPTIMIZER_MAP
        }
        build_inference_learner_ops = {
            compiler: COMPILER_TO_INFERENCE_LEARNER_MAP[compiler][
                self.pipeline_dl_framework
            ]()
            for compiler in compilers
            if compiler not in ignore_compilers
            and compiler in COMPILER_TO_OPTIMIZER_MAP
        }

        return compiler_ops, build_inference_learner_ops

    def _optimize(
        self,
        model: Union[torch.nn.Module, tf.Module, str],
        input_data: DataManager,
        optimization_time: OptimizationTime,
        metric_drop_ths: str,
        metric: Callable,
        model_params: ModelParams,
        model_outputs: List[Tuple[Any, ...]],
        ignore_compilers: List[ModelCompiler],
    ):

        if metric_drop_ths is not None:
            q_types = [
                None,
            ]
            if metric_drop_ths > 0:
                q_types.append(QuantizationType.HALF)
            if metric_drop_ths > ACTIVATION_METRIC_DROP_THS:
                q_types.append(QuantizationType.DYNAMIC)
                if input_data is not None:
                    q_types.append(QuantizationType.STATIC)
        else:
            q_types = [None]

        optimization_info = []
        for compiler, compiler_op, build_inference_learner_op in zip(
            self.compiler_ops.keys(),
            self.compiler_ops.values(),
            self.build_inference_learner_ops.values(),
        ):
            for q_type in q_types:
                input_tfms = MultiStageTransformation([])

                with TemporaryDirectory() as tmp_dir:
                    try:
                        compiler_op.to(self.device).execute(
                            model=model,
                            input_data=input_data,
                            model_params=model_params,
                            metric_drop_ths=metric_drop_ths
                            if q_type is not None
                            else None,
                            quantization_type=q_type,
                            input_tfms=input_tfms,
                            onnx_output_path=tmp_dir,
                        )

                        compiled_model = compiler_op.get_result()
                        if compiled_model is not None:
                            build_inference_learner_op.to(self.device).execute(
                                model=compiled_model,
                                model_orig=compiler_op.model_orig
                                if hasattr(compiler_op, "model_orig")
                                else None,
                                model_params=model_params,
                                input_tfms=input_tfms,
                                source_dl_framework=self.source_dl_framework,
                            )
                            inference_learner = (
                                build_inference_learner_op.get_result()
                            )

                            if inference_learner is not None:
                                test_input_data, ys = input_data.get_split(
                                    "test"
                                ).get_list(with_ys=True)

                                self.validity_check_op.execute(
                                    inference_learner,
                                    test_input_data,
                                    model_outputs,
                                    metric_drop_ths
                                    if q_type is not None
                                    else CONSTRAINED_METRIC_DROP_THS,
                                    metric_func=metric
                                    if q_type is not None
                                    else compute_relative_difference,
                                    ys=ys,
                                )

                                if self.validity_check_op.valid:
                                    latency = compute_optimized_running_time(
                                        inference_learner, input_data
                                    )
                                    self.logger.info(
                                        f"Optimized model latency: {latency} "
                                        f"sec/iter"
                                    )

                                    if (
                                        compiler not in ignore_compilers
                                        and optimization_time
                                        is OptimizationTime.CONSTRAINED
                                    ):
                                        ignore_compilers.append(compiler)

                                    self.optimized_models.append(
                                        (
                                            inference_learner,
                                            latency,
                                            self.validity_check_op.measure_result,  # noqa: E501
                                        )
                                    )

                                    opt_info_dict = {
                                        "compiler": f"{self.pipeline_dl_framework.value}_{compiler.value}",  # noqa: E501
                                        "technique": q_type.value
                                        if q_type
                                        else "none",
                                        "latency": latency,
                                    }
                                    if (
                                        metric_drop_ths is not None
                                        and q_type is not None
                                    ):
                                        opt_info_dict[
                                            "metric_loss"
                                        ] = (
                                            self.validity_check_op.measure_result  # noqa: E501
                                        )
                                        opt_info_dict[
                                            "metric"
                                        ] = metric.__name__
                                    optimization_info.append(opt_info_dict)
                                else:
                                    self.logger.warning(
                                        "The optimized model will be "
                                        "discarded due to poor results "
                                        "obtained with the given metric."
                                    )
                    except Exception as ex:
                        self.logger.warning(
                            f"Optimization failed with "
                            f"{self.pipeline_dl_framework} "
                            f"interface of {compiler}. Got error {ex}. "
                            f"If possible the compilation will be re-scheduled"
                            f" with another interface. Please consult the "
                            f"documentation for further info or open an issue "
                            f"on GitHub for receiving assistance."
                        )
                        optimization_info.append(
                            {
                                "compiler": compiler.value,
                                "technique": q_type.value
                                if q_type
                                else "none",
                                "latency": -1,
                            }
                        )
        if self.feedback_collector is not None:
            self.feedback_collector.store_info(
                key="optimizations",
                value=optimization_info,
            )

    def get_result(self) -> List:
        return self.optimized_models


COMPILER_TO_OPTIMIZER_MAP: Dict[
    ModelCompiler, Dict[DeepLearningFramework, Type[Compiler]]
] = {
    ModelCompiler.TORCHSCRIPT: {
        DeepLearningFramework.PYTORCH: PytorchBackendCompiler
    },
    ModelCompiler.DEEPSPARSE: {
        DeepLearningFramework.PYTORCH: DeepSparseCompiler
    },
    ModelCompiler.INTEL_NEURAL_COMPRESSOR: {
        DeepLearningFramework.PYTORCH: IntelNeuralCompressorCompiler
    },
    ModelCompiler.TENSOR_RT: {
        DeepLearningFramework.PYTORCH: PyTorchTensorRTCompiler,
        DeepLearningFramework.NUMPY: ONNXTensorRTCompiler,
    },
    ModelCompiler.APACHE_TVM: {
        DeepLearningFramework.PYTORCH: PyTorchApacheTVMCompiler,
        DeepLearningFramework.NUMPY: ONNXApacheTVMCompiler,
    },
    ModelCompiler.ONNX_RUNTIME: {DeepLearningFramework.NUMPY: ONNXCompiler},
    ModelCompiler.OPENVINO: {DeepLearningFramework.NUMPY: OpenVINOCompiler},
    ModelCompiler.TFLITE: {
        DeepLearningFramework.TENSORFLOW: TFLiteBackendCompiler
    },
    ModelCompiler.XLA: {
        DeepLearningFramework.TENSORFLOW: TensorflowBackendCompiler
    },
}

COMPILER_TO_INFERENCE_LEARNER_MAP: Dict[
    ModelCompiler, Dict[DeepLearningFramework, Type[BuildInferenceLearner]]
] = {
    ModelCompiler.TORCHSCRIPT: {
        DeepLearningFramework.PYTORCH: PytorchBuildInferenceLearner
    },
    ModelCompiler.DEEPSPARSE: {
        DeepLearningFramework.PYTORCH: DeepSparseBuildInferenceLearner
    },
    ModelCompiler.INTEL_NEURAL_COMPRESSOR: {
        DeepLearningFramework.PYTORCH: IntelNeuralCompressorBuildInferenceLearner  # noqa: E501
    },
    ModelCompiler.TENSOR_RT: {
        DeepLearningFramework.PYTORCH: PyTorchTensorRTBuildInferenceLearner,
        DeepLearningFramework.NUMPY: ONNXTensorRTBuildInferenceLearner,
    },
    ModelCompiler.APACHE_TVM: {
        DeepLearningFramework.PYTORCH: PyTorchApacheTVMBuildInferenceLearner,
        DeepLearningFramework.NUMPY: ONNXApacheTVMBuildInferenceLearner,
    },
    ModelCompiler.ONNX_RUNTIME: {
        DeepLearningFramework.NUMPY: ONNXBuildInferenceLearner
    },
    ModelCompiler.OPENVINO: {
        DeepLearningFramework.NUMPY: OpenVINOBuildInferenceLearner
    },
    ModelCompiler.TFLITE: {
        DeepLearningFramework.TENSORFLOW: TFLiteBuildInferenceLearner
    },
    ModelCompiler.XLA: {
        DeepLearningFramework.TENSORFLOW: TensorflowBuildInferenceLearner
    },
}



import os
import platform
import subprocess
import sys
from abc import ABC
from pathlib import Path
from typing import List, Union

import cpuinfo
from loguru import logger

from nebullvm.config import (
    LIBRARIES_GPU,
)
from nebullvm.operations.optimizations.compilers.utils import (
    openvino_is_available,
    tensorrt_is_available,
    torch_tensorrt_is_available,
    deepsparse_is_available,
    intel_neural_compressor_is_available,
)
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.utils import (
    gpu_is_available,
    check_module_version,
)


def get_cpu_arch():
    arch = cpuinfo.get_cpu_info()["arch"].lower()
    if "x86" in arch:
        return "x86"
    else:
        return "arm"


def _get_os():
    return platform.system()


def install_tvm(
    working_dir: str = None,
):
    """Helper function for installing ApacheTVM.

    This function needs some prerequisites for running, as a valid `git`
    installation and having MacOS or a Linux-distribution as OS.

    Args:
        working_dir (str, optional): The directory where the tvm repo will be
            cloned and installed.
    """
    path = Path(__file__).parent
    # install pre-requisites
    installation_file_prerequisites = str(
        path / "install_tvm_prerequisites.sh"
    )
    subprocess.run(
        ["bash", installation_file_prerequisites],
        cwd=working_dir or Path.home(),
    )
    installation_file = str(path / "install_tvm.sh")
    hardware_config = get_cpu_arch()
    if gpu_is_available():
        hardware_config = f"{hardware_config}_cuda"
    env_dict = {
        "CONFIG_PATH": str(
            path / f"tvm_installers/{hardware_config}/config.cmake"
        ),
        **dict(os.environ.copy()),
    }
    subprocess.run(
        ["bash", installation_file],
        cwd=working_dir or Path.home(),
        env=env_dict,
    )

    try:
        import tvm  # noqa F401
    except ImportError:
        return True

    return True


def install_bladedisc():
    """Helper function for installing BladeDisc."""
    has_cuda = False
    if gpu_is_available():
        has_cuda = True

    path = Path(__file__).parent
    installation_file = str(path / "install_bladedisc.sh")
    subprocess.Popen(["bash", installation_file, str(has_cuda).lower()])

    try:
        import torch_blade  # noqa F401
    except ImportError:
        return False

    return True


def install_torch_tensor_rt():
    """Helper function for installing Torch-TensorRT.

    The function will install the software only if a cuda driver is available.
    """
    if not gpu_is_available():
        raise RuntimeError(
            "Torch-TensorRT can run just on Nvidia machines. "
            "No available cuda driver has been found."
        )
    elif not check_module_version(torch, min_version="1.12.0"):
        raise RuntimeError(
            "Torch-TensorRT can be installed only from Pytorch 1.12. "
            "Please update your Pytorch version."
        )

    # Verify that TensorRT is installed, otherwise install it
    try:
        import tensorrt  # noqa F401
    except ImportError:
        install_tensor_rt()

    # cmd = [
    #     "pip3",
    #     "install",
    #     "torch-tensorrt>=1.2.0",
    #     "-f",
    #     "https://github.com/pytorch/TensorRT/releases",
    # ]

    cmd = [
        "pip3",
        "install",
        "torch-tensorrt",
        "--find-links",
        "https://github.com/pytorch/TensorRT/releases/expanded_assets/v1.3.0",
    ]
    subprocess.run(cmd)

    try:
        import torch_tensorrt  # noqa F401
    except ImportError:
        return False

    return True


def install_tf2onnx():
    if _get_os() == "Darwin" and get_cpu_arch() == "arm":
        cmd = ["conda", "install", "-y", "tf2onnx>=1.8.4"]
        subprocess.run(cmd)
    else:
        cmd = ["pip3", "install", "--user", "protobuf<4,>=3.20.2"]
        subprocess.run(cmd)

        cmd = ["pip3", "install", "tf2onnx>=1.8.4"]
        subprocess.run(cmd)

    try:
        import tf2onnx  # noqa F401
    except ImportError:
        return False
    except AttributeError:
        # Sometimes the import could raise an attribute error
        # if installation fails
        pass

    return True


def install_tensor_rt():
    """Helper function for installing TensorRT.

    The function will install the software only if a cuda driver is available.
    """
    if not gpu_is_available():
        raise RuntimeError(
            "TensorRT can run just on Nvidia machines. "
            "No available cuda driver has been found."
        )
    path = Path(__file__).parent
    installation_file = str(path / "install_tensor_rt.sh")
    subprocess.run(["bash", installation_file])

    try:
        import polygraphy  # noqa F401
        import tensorrt  # noqa F401
    except ImportError:
        return False

    return True


def install_openvino(with_optimization: bool = True):
    """Helper function for installing the OpenVino compiler.

    This function just works on intel machines.

    Args:
        with_optimization (bool): Flag for installing the full openvino engine
            or limiting the installation to the tools need for inference
            models.
    """
    processor = cpuinfo.get_cpu_info()["brand_raw"].lower()
    if "intel" not in processor:
        raise RuntimeError(
            f"Openvino can run just on Intel machines. "
            f"You are trying to install it on {processor}"
        )

    openvino_version = "openvino-dev" if with_optimization else "openvino"
    cmd = ["pip3", "install", f"{openvino_version}>=2022.1.0"]
    subprocess.run(cmd)

    cmd = ["pip3", "install", "scipy>=1.7.3"]
    subprocess.run(cmd)

    try:
        from openvino.runtime import (  # noqa F401
            Core,
            Model,
            CompiledModel,
            InferRequest,
        )
    except ImportError:
        return False

    return True


def install_onnxruntime():
    """Helper function for installing the right version of onnxruntime."""
    distribution_name = "onnxruntime"
    if gpu_is_available():
        distribution_name = f"{distribution_name}-gpu"
    if _get_os() == "Darwin" and get_cpu_arch() == "arm":
        cmd = ["conda", "install", "-y", distribution_name]
    else:
        cmd = ["pip3", "install", distribution_name]
    subprocess.run(cmd)
    # install requirements for onnxruntime.transformers
    cmd = ["pip3", "install", "coloredlogs", "sympy"]
    subprocess.run(cmd)

    try:
        import onnxruntime  # noqa F401
    except ImportError:
        return False

    return True


def install_deepsparse():
    """Helper function for installing DeepSparse."""
    python_minor_version = sys.version_info.minor

    os_ = platform.system()
    if os_ in ["Darwin", "Windows"] or get_cpu_arch() == "arm":
        raise RuntimeError(
            "DeepSparse is not supported on this platform. "
            "It won't be installed."
        )

    try:
        cmd = ["apt-get", "install", f"python3.{python_minor_version}-venv"]
        subprocess.run(cmd)
    except Exception:
        pass

    cmd = ["pip3", "install", "deepsparse"]
    subprocess.run(cmd)

    try:
        from deepsparse import compile_model, cpu  # noqa F401
    except ImportError:
        return False

    return True


def install_intel_neural_compressor():
    """Helper function for installing Intel Neural Compressor."""

    processor = cpuinfo.get_cpu_info()["brand_raw"].lower()
    if "intel" not in processor:
        raise RuntimeError(
            f"Intel Neural Compressor can run just on Intel machines. "
            f"You are trying to install it on {processor}"
        )

    cmd = ["pip3", "install", "--user", "neural-compressor"]
    subprocess.run(cmd)

    try:
        from neural_compressor.experimental import (  # noqa F401
            MixedPrecision,
            Quantization,
        )
    except ImportError:
        return False

    return True


def install_onnx_simplifier():
    """Helper function for installing ONNX simplifier."""

    if get_cpu_arch() != "arm":
        # Install onnx simplifier
        cmd = ["pip3", "install", "onnxsim"]
        subprocess.run(cmd)

    try:
        import onnxsim  # noqa F401
    except ImportError:
        return False

    return True


class BaseInstaller(ABC):
    def __init__(self, module_list: List[str]):
        self.modules = module_list

    def install_compilers(
        self,
        include_libraries: Union[List[str], str] = "all",
    ):
        for library in self.modules:
            if (
                isinstance(include_libraries, List)
                and library not in include_libraries
            ) or (not gpu_is_available() and library in LIBRARIES_GPU):
                continue

            logger.info(f"Trying to install {library} on the platform...")

            try:
                if not COMPILERS_AVAILABLE[library]():
                    install_ok = COMPILER_INSTALLERS[library]()
                else:
                    install_ok = True
            except Exception:
                install_ok = False

            if not install_ok:
                logger.warning(
                    f"Unable to install {library} on this platform. "
                    f"The compiler will be skipped. "
                )
            else:
                logger.info(f"{library} installed successfully!")

    @staticmethod
    def install_dependencies(include_backend: List[str]):
        raise NotImplementedError

    @staticmethod
    def check_backend():
        raise NotImplementedError

    @staticmethod
    def install_backend():
        raise NotImplementedError


class PytorchInstaller(BaseInstaller):
    @staticmethod
    def install_dependencies(include_backend: List[str]):
        return

    @staticmethod
    def check_backend():
        try:
            import torch  # noqa F401
        except ImportError:
            raise ImportError(
                "No PyTorch found in your python environment. Please install "
                "it from https://pytorch.org/get-started/locally/."
            )

        if not check_module_version(torch, min_version="1.12.0"):
            logger.warning(
                "You are using an old version of PyTorch, please update it "
                "in order to get the best optimization results."
            )

        return True

    @staticmethod
    def install_backend():
        cmd = ["pip3", "install", "torch>=1.10.0"]
        subprocess.run(cmd)

        try:
            import torch  # noqa F401
        except ImportError:
            return False

        return True


class TensorflowInstaller(BaseInstaller):
    @staticmethod
    def install_dependencies(include_backend: List[str]):
        if "onnx" in include_backend:
            install_tf2onnx()

    @staticmethod
    def check_backend():
        try:
            import tensorflow  # noqa F401
        except ImportError:
            return False

        if not check_module_version(tensorflow, min_version="2.7.0"):
            return False

        return True

    @staticmethod
    def install_backend():
        if _get_os() == "Darwin" and get_cpu_arch() == "arm":
            cmd = ["conda", "install", "-y", "tensorflow>=2.7.0", "numpy<1.24"]
            subprocess.run(cmd)
        else:
            cmd = ["pip3", "install", "--user", "tensorflow>=2.7.0"]
            subprocess.run(cmd)

        try:
            import tensorflow  # noqa F401
        except ImportError:
            return False

        return True


class ONNXInstaller(BaseInstaller):
    @staticmethod
    def install_dependencies(include_backend: List[str]):
        install_onnxruntime()
        cmd = ["pip3", "install", "onnxmltools>=1.11.0"]
        subprocess.run(cmd)
        install_onnx_simplifier()

    @staticmethod
    def check_backend():
        try:
            import onnx  # noqa F401
        except ImportError:
            return False

        if not check_module_version(onnx, min_version="1.10.0"):
            return False

        return True

    @staticmethod
    def install_backend():
        if _get_os() == "Darwin" and get_cpu_arch() == "arm":
            cmd = ["pip3", "install", "cmake"]
            subprocess.run(cmd)

        cmd = ["pip3", "install", "onnx>=1.10.0"]
        subprocess.run(cmd)

        try:
            import onnx  # noqa F401
        except ImportError:
            return False

        return True


class HuggingFaceInstaller(BaseInstaller):
    @staticmethod
    def install_dependencies(include_backend: List[str]):
        pass

    @staticmethod
    def check_backend():
        try:
            import transformers  # noqa F401
        except ImportError:
            return False

        return True

    @staticmethod
    def install_backend():
        cmd = ["pip3", "install", "transformers"]
        subprocess.run(cmd)

        try:
            import transformers  # noqa F401
        except ImportError:
            return False

        return True


COMPILER_INSTALLERS = {
    "openvino": install_openvino,
    "tensor_rt": install_tensor_rt,
    "torch_tensor_rt": install_torch_tensor_rt,
    "deepsparse": install_deepsparse,
    "intel_neural_compressor": install_intel_neural_compressor,
}

COMPILERS_AVAILABLE = {
    "openvino": openvino_is_available,
    "tensor_rt": tensorrt_is_available,
    "torch_tensor_rt": torch_tensorrt_is_available,
    "deepsparse": deepsparse_is_available,
    "intel_neural_compressor": intel_neural_compressor_is_available,
}



from collections import OrderedDict
from typing import (
    Union,
    Iterable,
    List,
    Dict,
    Tuple,
    Type,
    Any,
)

import numpy as np

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch, Module
from nebullvm.tools.base import Device

try:
    from transformers import (
        PreTrainedModel,
    )
    from transformers.tokenization_utils import PreTrainedTokenizer
except ImportError:
    # add placeholders for function definition
    PreTrainedModel = None
    PreTrainedTokenizer = None


class PyTorchTransformerWrapper(Module):
    """Class for wrappering the Transformers and give them an API compatible
    with nebullvm. The class takes and input of the forward method positional
    arguments and transform them in the input dictionaries needed by
    transformers classes. At the end it also flattens their output.
    """

    def __init__(
        self,
        core_model: Module,
        encoded_input: Dict[str, torch.Tensor],
    ):
        super().__init__()
        self.core_model = core_model
        self.inputs_types = OrderedDict()
        for key, value in encoded_input.items():
            self.inputs_types[key] = value.dtype

    def forward(self, *args: torch.Tensor):
        inputs = {
            key: value for key, value in zip(self.inputs_types.keys(), args)
        }
        outputs = self.core_model(**inputs)
        outputs = outputs.values() if isinstance(outputs, dict) else outputs
        return tuple(flatten_outputs(outputs))


class TensorFlowTransformerWrapper(tf.keras.Model):
    def __init__(
        self,
        core_model: tf.Module,
        encoded_input: Dict[str, tf.Tensor],
    ):
        super().__init__()
        self.core_model = core_model
        self.inputs_types = OrderedDict()
        for key, value in encoded_input.items():
            self.inputs_types[key] = value.dtype

    def call(self, *args: tf.Tensor):
        inputs = {
            key: value for key, value in zip(self.inputs_types.keys(), args[0])
        }
        outputs = self.core_model(**inputs)
        outputs = outputs.values() if isinstance(outputs, dict) else outputs
        return tuple(flatten_outputs(list(outputs)))


def flatten_outputs(
    outputs: Union[torch.Tensor, tf.Tensor, Iterable]
) -> List[Union[torch.Tensor, tf.Tensor]]:
    new_outputs = []
    for output in outputs:
        if isinstance(output, (torch.Tensor, tf.Tensor)):
            new_outputs.append(output)
        else:
            flatten_list = flatten_outputs(output)
            new_outputs.extend(flatten_list)
    return new_outputs


def get_size_recursively(
    tensor_tuple: Union[torch.Tensor, tf.Tensor, Tuple]
) -> List[int]:
    if isinstance(tensor_tuple[0], (torch.Tensor, tf.Tensor)):
        return [len(tensor_tuple)]
    else:
        inner_size = get_size_recursively(tensor_tuple[0])
        return [len(tensor_tuple), *inner_size]


def get_output_structure_from_text(
    text: str,
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizer,
    tokenizer_args: Dict,
    device: Device,
) -> Tuple[OrderedDict, Type]:
    """Function needed for saving in a dictionary the output structure of the
    transformers model.
    """
    encoded_input = tokenizer([text], **tokenizer_args)
    if isinstance(model, torch.nn.Module):
        device = torch.device("cuda" if device is Device.GPU else "cpu")
        encoded_input = encoded_input.to(device)
    output = model(**encoded_input)
    structure = OrderedDict()
    if isinstance(output, tuple):
        for i, value in enumerate(output):
            if isinstance(value, (torch.Tensor, tf.Tensor)):
                structure[f"output_{i}"] = None
            else:
                size = get_size_recursively(value)
                structure[f"output_{i}"] = size
    else:
        for key, value in output.items():
            if isinstance(value, (torch.Tensor, tf.Tensor)):
                structure[key] = None
            else:
                size = get_size_recursively(value)
                structure[key] = size
    return structure, type(output)


def get_output_structure_from_dict(
    input_example: Dict,
    model: PreTrainedModel,
    device: Device,
) -> Tuple[OrderedDict, Type]:
    """Function needed for saving in a dictionary the output structure of the
    transformers model.
    """

    if isinstance(model, torch.nn.Module):
        device = torch.device("cuda" if device is Device.GPU else "cpu")
        model.to(device)
        input_example.to(device)

    output = model(**input_example)
    structure = OrderedDict()
    if isinstance(output, tuple):
        for i, value in enumerate(output):
            if isinstance(value, (torch.Tensor, tf.Tensor)):
                structure[f"output_{i}"] = None
            else:
                size = get_size_recursively(value)
                structure[f"output_{i}"] = size
    else:
        for key, value in output.items():
            if isinstance(value, (torch.Tensor, tf.Tensor)):
                structure[key] = None
            else:
                size = get_size_recursively(value)
                structure[key] = size
    return structure, type(output)


def restructure_output(
    output: Tuple[Union[torch.Tensor, tf.Tensor]],
    structure: OrderedDict,
    output_type: Any = None,
):
    """Restructure the flatter output using the structure dictionary given as
    input.
    """
    output_dict = {}
    idx = 0
    for key, value in structure.items():
        if value is None:
            output_dict[key] = output[idx]
            idx += 1
        else:
            tensor_shape = output[idx].shape[1:]
            stack_fn = (
                torch.stack
                if isinstance(output[idx], torch.Tensor)
                else tf.stack
            )
            reshape_fn = (
                torch.reshape
                if isinstance(output[idx], torch.Tensor)
                else tf.reshape
            )

            output_dict[key] = list(
                reshape_fn(
                    stack_fn(
                        output[idx : int(np.prod(value)) + idx]  # noqa E203
                    ),
                    (*value, *tensor_shape),
                )
            )
            idx += np.prod(value)
    if output_type is not None:
        return output_type(**output_dict)
    return output_dict



try:
    import tensorrt
    from tensorrt import IInt8EntropyCalibrator2
except ImportError:
    tensorrt = object
    IInt8EntropyCalibrator2 = object

try:
    import polygraphy.cuda as polygraphy
except ImportError:
    polygraphy = object



from pathlib import Path
from typing import Union, Any

from nebullvm.operations.inference_learners.base import BuildInferenceLearner
from nebullvm.operations.inference_learners.deepsparse import (
    PytorchDeepSparseInferenceLearner,
)
from nebullvm.operations.inference_learners.neural_compressor import (
    PytorchNeuralCompressorInferenceLearner,
)
from nebullvm.operations.inference_learners.onnx import ONNX_INFERENCE_LEARNERS
from nebullvm.operations.inference_learners.openvino import (
    OPENVINO_INFERENCE_LEARNERS,
)
from nebullvm.operations.inference_learners.pytorch import (
    PytorchBackendInferenceLearner,
)
from nebullvm.operations.inference_learners.tensor_rt import (
    PytorchTensorRTInferenceLearner,
    TENSOR_RT_INFERENCE_LEARNERS,
)
from nebullvm.operations.inference_learners.tensorflow import (
    TensorflowBackendInferenceLearner,
    TFLiteBackendInferenceLearner,
)
from nebullvm.operations.inference_learners.tvm import (
    PytorchApacheTVMInferenceLearner,
    APACHE_TVM_INFERENCE_LEARNERS,
)
from nebullvm.optional_modules.tensor_rt import tensorrt as trt
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import ScriptModule, Module, GraphModule
from nebullvm.optional_modules.tvm import tvm, ExecutorFactoryModule
from nebullvm.tools.base import DeepLearningFramework, ModelParams, Device
from nebullvm.tools.onnx import get_input_names, get_output_names
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    VerifyContiguity,
)


class PytorchBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: ScriptModule,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        **kwargs,
    ):
        self.inference_learner = PytorchBackendInferenceLearner(
            torch_model=model,
            network_parameters=model_params,
            input_tfms=input_tfms,
            device=self.device,
        )


class TensorflowBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: tf.Module,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        **kwargs,
    ):
        self.inference_learner = TensorflowBackendInferenceLearner(
            model,
            network_parameters=model_params,
            input_tfms=input_tfms,
            device=self.device,
        )


class TFLiteBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: bytes,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        **kwargs,
    ):
        self.inference_learner = TFLiteBackendInferenceLearner(
            model,
            network_parameters=model_params,
            input_tfms=input_tfms,
            device=self.device,
        )


class DeepSparseBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: Union[str, Path],
        model_params: ModelParams,
        **kwargs,
    ):
        input_names = get_input_names(str(model))
        output_names = get_output_names(str(model))

        self.inference_learner = PytorchDeepSparseInferenceLearner(
            onnx_path=model,
            network_parameters=model_params,
            input_names=input_names,
            output_names=output_names,
        )


class ONNXBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: Union[str, Path],
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        source_dl_framework: DeepLearningFramework,
        **kwargs,
    ):
        input_names = get_input_names(str(model))
        output_names = get_output_names(str(model))

        self.inference_learner = ONNX_INFERENCE_LEARNERS[source_dl_framework](
            onnx_path=model,
            network_parameters=model_params,
            input_names=input_names,
            output_names=output_names,
            input_tfms=input_tfms,
            device=self.device,
        )


class OpenVINOBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: str,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        source_dl_framework: DeepLearningFramework,
        **kwargs,
    ):
        self.inference_learner = OPENVINO_INFERENCE_LEARNERS[
            source_dl_framework
        ].from_model_name(
            model_name=model + ".xml",
            model_weights=model + ".bin",
            input_tfms=input_tfms,
            network_parameters=model_params,
        )


class PyTorchTensorRTBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: ScriptModule,
        input_tfms: MultiStageTransformation,
        model_params: ModelParams,
        **kwargs,
    ):
        self.inference_learner = PytorchTensorRTInferenceLearner(
            torch_model=model,
            input_tfms=input_tfms,
            network_parameters=model_params,
            device=self.device,
        )


class ONNXTensorRTBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: Any,
        model_orig: Union[str, Path],
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        source_dl_framework: DeepLearningFramework,
        **kwargs,
    ):
        nvidia_logger = trt.Logger(trt.Logger.ERROR)
        input_names = get_input_names(str(model_orig))
        output_names = get_output_names(str(model_orig))

        input_tfms.append(VerifyContiguity())
        runtime = trt.Runtime(nvidia_logger)
        engine = runtime.deserialize_cuda_engine(model)

        self.inference_learner = TENSOR_RT_INFERENCE_LEARNERS[
            source_dl_framework
        ](
            engine=engine,
            input_tfms=input_tfms,
            network_parameters=model_params,
            input_names=input_names,
            output_names=output_names,
            nvidia_logger=nvidia_logger,
            device=self.device,
        )


class IntelNeuralCompressorBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: GraphModule,
        model_orig: Module,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        **kwargs,
    ):
        self.inference_learner = PytorchNeuralCompressorInferenceLearner(
            model=model_orig,
            model_quant=model,
            input_tfms=input_tfms,
            network_parameters=model_params,
            device=self.device,
        )


class PyTorchApacheTVMBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: ExecutorFactoryModule,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        **kwargs,
    ):
        target_device = (
            str(tvm.target.cuda()) if self.device is Device.GPU else "llvm"
        )
        dev = tvm.device(str(target_device), 0)

        input_names = [
            f"input_{i}" for i in range(len(model_params.input_infos))
        ]

        graph_executor_module = tvm.contrib.graph_executor.GraphModule(
            model["default"](dev)
        )
        self.inference_learner = PytorchApacheTVMInferenceLearner(
            input_tfms=input_tfms,
            network_parameters=model_params,
            graph_executor_module=graph_executor_module,
            input_names=input_names,
            lib=model,
            target=target_device,
        )


class ONNXApacheTVMBuildInferenceLearner(BuildInferenceLearner):
    def execute(
        self,
        model: ExecutorFactoryModule,
        model_orig: str,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        source_dl_framework: DeepLearningFramework,
        **kwargs,
    ):
        target_device = (
            str(tvm.target.cuda()) if self.device is Device.GPU else "llvm"
        )
        dev = tvm.device(str(target_device), 0)

        input_names = (
            get_input_names(model_orig)
            if model_orig is not None
            else [f"input_{i}" for i in range(len(model_params.input_infos))]
        )

        graph_executor_module = tvm.contrib.graph_executor.GraphModule(
            model["default"](dev)
        )
        self.inference_learner = APACHE_TVM_INFERENCE_LEARNERS[
            source_dl_framework
        ](
            input_tfms=input_tfms,
            network_parameters=model_params,
            graph_executor_module=graph_executor_module,
            input_names=input_names,
            lib=model,
            target=target_device,
        )



class Onnxruntime:
    pass


try:
    import onnxruntime  # noqa F401
    from onnxruntime.quantization import (
        QuantType,
        quantize_static,
        quantize_dynamic,
        CalibrationDataReader,
    )
except ImportError:
    onnxruntime = Onnxruntime
    setattr(onnxruntime, "SessionOptions", None)
    QuantType = quantize_static = quantize_dynamic = None
    CalibrationDataReader = object
except FileNotFoundError:
    # Solves a colab issue
    QuantType = quantize_static = quantize_dynamic = None
    CalibrationDataReader = object

try:
    # They require torch
    from onnxruntime.transformers import optimizer
    from onnxruntime.transformers.optimizer import MODEL_TYPES
except ImportError:
    MODEL_TYPES = object
    optimizer = object



import abc
from typing import Any, Dict, List, Optional

from nebullvm.operations.base import Operation
from nebullvm.tools.base import QuantizationType


class Compiler(Operation, abc.ABC):
    supported_ops: Dict[str, List[Optional[QuantizationType]]]

    def __init__(self):
        super().__init__()
        self.compiled_model = None

    @abc.abstractmethod
    def execute(self, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compile_model(self, **kwargs) -> Any:
        raise NotImplementedError()

    @abc.abstractmethod
    def _quantize_model(self, **kwargs) -> Any:
        raise NotImplementedError()

    def get_result(self) -> Any:
        return self.compiled_model



import json
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Dict

import numpy as np
from loguru import logger

from nebullvm.operations.optimizations.compressors.base import Compressor
from nebullvm.optional_modules.torch import torch, Module
from nebullvm.tools.data import DataManager
from nebullvm.tools.pytorch import save_with_torch_fx, load_with_torch_fx
from nebullvm.tools.venv import run_in_different_venv


def _save_model(model: Module, path: Path):
    try:
        save_with_torch_fx(model, path)
    except Exception as ex:
        logger.warning(
            f"Got an error while exporting with TorchFX. The model will be "
            f"saved using the standard PyTorch save pickling method. Error "
            f"got: {ex}"
        )
        torch.save(model, path / "model.pt")
        return path / "model.pt"
    else:
        return path


def _load_model(path: Path):
    if path.is_file():
        return torch.load(path)
    else:
        return load_with_torch_fx(path)


def _save_dataset(input_data: DataManager, path: Path):
    path.mkdir(exist_ok=True)
    for i, x in enumerate(input_data):
        torch.save(x, path / f"input_{i}.pt")


def _save_json(dictionary: Dict, path: Path):
    with open(path, "w") as f:
        json.dump(dictionary, f)


def _write_requirements_file(path: Path):
    requirements = "sparseml\nsparsify\ntqdm"
    with open(path, "w") as f:
        f.write(requirements)


class SparseMLCompressor(Compressor):
    def execute(
        self,
        model: Module,
        train_input_data: DataManager,
        eval_input_data: DataManager,
        metric_drop_ths: float,
        metric: Callable,
    ):
        script_path = (
            Path(__file__).parent / "scripts/neural_magic_training.py"
        )
        with TemporaryDirectory(dir=".") as tmp_dir:
            tmp_dir = Path(tmp_dir)
            requirements_file = tmp_dir / "requirements.txt"
            model_path = _save_model(model, tmp_dir)
            training_data_dir = tmp_dir / "train"
            eval_data_dir = tmp_dir / "eval"
            config_file = tmp_dir / "config.json"
            pruned_model_path = (
                tmp_dir / "pruned_model.pt"
                if model_path.is_file()
                else tmp_dir
            )

            _write_requirements_file(requirements_file)
            _save_dataset(train_input_data, training_data_dir)
            _save_dataset(eval_input_data, eval_data_dir)
            _save_json(self._config, config_file)

            run_in_different_venv(
                str(requirements_file),
                str(script_path),
                torch.cuda.is_available(),
                "--model",
                f"{model_path}",
                "--train_dir",
                f"{training_data_dir}",
                "--eval_dir",
                f"{eval_data_dir}",
                "--config",
                f"{config_file}",
                "--pruned_model",
                f"{pruned_model_path}",
            )

            self.compressed_model = _load_model(pruned_model_path)

            if self.compressed_model is not None:
                error = self._compute_error(
                    model, self.compressed_model, eval_input_data, metric
                )
                if error > metric_drop_ths:
                    self.compressed_model = None
                else:
                    self.new_metric_ths = metric_drop_ths - error

    @staticmethod
    @torch.no_grad()
    def _compute_error(
        model: Module,
        pruned_model: Module,
        eval_input_data: DataManager,
        metric: Callable,
    ) -> float:
        if len(eval_input_data) == 0:
            return np.inf
        metric_val = 0.0
        model.eval()
        pruned_model.eval()
        for inputs, y in eval_input_data:
            if torch.cuda.is_available():
                inputs = tuple(data.cuda() for data in inputs)
                pruned_model.cuda()
                model.cuda()
            model_pred = model(*inputs)
            pruned_pred = pruned_model(*inputs)
            metric_val += metric(model_pred, pruned_pred, y)
        return metric_val / len(eval_input_data)

    @staticmethod
    def _get_default_config() -> Dict:
        return {
            "training_epochs": 10,
            "epochs_pruning_window": {"start_epoch": 0, "end_epoch": 10},
            "loss_fn": "CrossEntropy",
            "lr": 1e-3,
            "momentum": 0.9,
        }

    @property
    def config_key(self) -> str:
        return "sparseml"



import abc
import os
import uuid
from abc import ABC
from typing import Any, Tuple, Dict, Union

from nebullvm.config import (
    AUTO_TVM_PARAMS,
    AUTO_TVM_TUNING_OPTION,
)
from nebullvm.operations.optimizations.compilers.base import Compiler
from nebullvm.operations.optimizations.compilers.quantizations.tvm import (
    TVMCalibrator,
    quantize_apache_tvm,
)
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.optional_modules.onnx import onnx
from nebullvm.optional_modules.torch import Module, torch
from nebullvm.optional_modules.tvm import (
    tvm,
    IRModule,
    NDArray,
    XGBTuner,
    autotvm,
    relay,
    ExecutorFactoryModule,
)
from nebullvm.tools.base import QuantizationType, ModelParams, Device
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import get_input_names
from nebullvm.tools.pytorch import create_model_inputs_torch
from nebullvm.tools.transformations import MultiStageTransformation


class ApacheTVMCompiler(Compiler, ABC):
    supported_ops = {
        "cpu": [
            None,
            # QuantizationType.STATIC,
            QuantizationType.HALF,
            QuantizationType.DYNAMIC,
        ],
        "gpu": [
            None,
            # QuantizationType.STATIC,
            QuantizationType.HALF,
            QuantizationType.DYNAMIC,
        ],
    }

    def __init__(self):
        super().__init__()
        self.model_orig = None

    def execute(
        self,
        model: Union[Module, str],
        input_tfms: MultiStageTransformation,
        model_params: ModelParams,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Compile the input model using Apache TVM compiler.

        Args:
            model (Union[Module, str]: The input model. Can be a torch model
                or a path to an onnx model.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            model_params (ModelParams): Model parameters.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)

        mod, params = self._build_tvm_model(model, model_params)

        if quantization_type is not None:
            mod = self._quantize_model(
                mod, quantization_type, input_tfms, input_data, params
            )

        self.compiled_model = self._compile_model(mod, params)

    @abc.abstractmethod
    def _build_tvm_model(self, model: Any, model_params: ModelParams):
        raise NotImplementedError()

    @staticmethod
    def _build_tvm_model_from_torch(
        torch_model: Module, model_params: ModelParams, device: Device
    ) -> Tuple[IRModule, Dict[str, NDArray]]:
        shape_dict = {
            f"input_{i}": (
                model_params.batch_size,
                *input_size,
            )
            for i, input_size in enumerate(model_params.input_sizes)
        }
        inputs = tuple(
            create_model_inputs_torch(
                model_params.batch_size, model_params.input_infos
            )
        )
        if device is not Device.GPU:
            inputs = tuple(input_.cpu() for input_ in inputs)
            torch_model.cpu()
        else:
            inputs = tuple(input_.cuda() for input_ in inputs)
            torch_model.cuda()
        with torch.no_grad():
            _ = torch_model(*inputs)
            model_trace = torch.jit.trace(torch_model, inputs)
            model_trace.eval()
        mod, params = relay.frontend.from_pytorch(
            model_trace, list(shape_dict.items())
        )
        return mod, params

    @staticmethod
    def _build_tvm_model_from_onnx(
        onnx_model_path: str, model_params: ModelParams
    ) -> Tuple[IRModule, Dict[str, NDArray]]:
        shape_dict = {
            input_key: (
                model_params.batch_size,
                *input_size,
            )
            for input_key, input_size in zip(
                get_input_names(onnx_model_path), model_params.input_sizes
            )
        }
        onnx_model = onnx.load(onnx_model_path)
        mod, params = relay.frontend.from_onnx(onnx_model, shape_dict)
        return mod, params

    @staticmethod
    def _quantize(
        mod: IRModule,
        params: Dict[str, NDArray],
        input_data: TVMCalibrator = None,
    ) -> IRModule:
        if input_data is not None:
            with relay.quantize.qconfig(
                calibrate_mode="kl_divergence", weight_scale="max"
            ):
                mod = relay.quantize.quantize(mod, params, dataset=input_data)
        else:
            with relay.quantize.qconfig(
                calibrate_mode="global_scale", global_scale=8.0
            ):
                mod = relay.quantize.quantize(mod, params)
        return mod

    @staticmethod
    def _get_target(device) -> str:
        if device is Device.GPU:
            return str(tvm.target.cuda())
        else:
            return "llvm"  # run on CPU

    @staticmethod
    def _tune_tvm_model(
        target: str, mod: IRModule, params: Dict[str, NDArray]
    ) -> str:
        """Tune the model using AutoTVM."""
        # TODO: add support to Ansor
        tuning_records = f"{uuid.uuid4()}_model_records.json"
        # create a TVM runner
        runner = autotvm.LocalRunner(
            number=AUTO_TVM_PARAMS["number"],
            repeat=AUTO_TVM_PARAMS["repeat"],
            timeout=AUTO_TVM_PARAMS["timeout"],
            min_repeat_ms=AUTO_TVM_PARAMS["min_repeat_ms"],
            # TODO modify min_repeat_ms for GPU usage
            enable_cpu_cache_flush=True,
        )
        # begin by extracting the tasks from the onnx model
        tasks = autotvm.task.extract_from_program(
            mod["main"], target=target, params=params
        )

        # Tune the extracted tasks sequentially.
        for i, task in enumerate(tasks):
            tuner_obj = XGBTuner(task, loss_type="rank")
            tuner_obj.tune(
                n_trial=min(
                    AUTO_TVM_TUNING_OPTION["trials"], len(task.config_space)
                ),
                early_stopping=AUTO_TVM_TUNING_OPTION["early_stopping"],
                measure_option=autotvm.measure_option(
                    builder=autotvm.LocalBuilder(build_func="default"),
                    runner=runner,
                ),
                callbacks=[
                    autotvm.callback.log_to_file(tuning_records),
                ],
            )
        return tuning_records

    def _compile_model(self, model: Any, params: Any) -> ExecutorFactoryModule:
        target = self._get_target(self.device)
        tuning_records = self._tune_tvm_model(target, model, params)
        with autotvm.apply_history_best(tuning_records):
            with tvm.transform.PassContext(opt_level=3, config={}):
                lib = relay.build(model, target=target, params=params)

        # Remove temporary file created by tvm
        os.remove(tuning_records)

        return lib

    @staticmethod
    def _quantize_model(
        model: Any,
        quantization_type: QuantizationType,
        input_tfms: MultiStageTransformation,
        input_data: DataManager,
        params,
    ):
        return quantize_apache_tvm(
            model, quantization_type, input_tfms, input_data, params
        )


class PyTorchApacheTVMCompiler(ApacheTVMCompiler):
    def _build_tvm_model(self, model: Any, model_params: ModelParams):
        return self._build_tvm_model_from_torch(
            model, model_params, self.device
        )


class ONNXApacheTVMCompiler(ApacheTVMCompiler):
    def _build_tvm_model(self, model: Any, model_params: ModelParams):
        self.model_orig = model
        return self._build_tvm_model_from_onnx(model, model_params)



import os
import pickle
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple, Union, Optional, List

from nebullvm.operations.inference_learners.base import (
    PytorchBaseInferenceLearner,
    LearnerMetadata,
)
from nebullvm.optional_modules.torch import (
    torch,
    symbolic_trace,
    Module,
    ScriptModule,
    GraphModule,
)
from nebullvm.tools.base import Device, ModelParams
from nebullvm.tools.transformations import MultiStageTransformation


class PytorchBackendInferenceLearner(PytorchBaseInferenceLearner):
    MODEL_NAME = "model_scripted.pt"
    name = "TorchScript"

    def __init__(self, torch_model: ScriptModule, device: Device, **kwargs):
        super().__init__(**kwargs)
        self.model = torch_model.eval()
        if device is Device.GPU:
            self.model.cuda()
        self.device = device

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        device = input_tensors[0].device
        if self.device is Device.GPU:
            input_tensors = (t.cuda() for t in input_tensors)
        with torch.no_grad():
            res = self.model(*input_tensors)
            if not isinstance(res, tuple):
                res = res.to(device)
                return (res,)
            return tuple(out.to(device) for out in res)

    def get_size(self):
        try:
            if hasattr(self.model, "core_model"):
                return len(pickle.dumps(self.model.core_model, -1))
            else:
                # Normal torch model
                return len(pickle.dumps(self.model, -1))
        except RuntimeError:
            with TemporaryDirectory() as tmp_dir:
                self.save(tmp_dir)
                return sum(
                    os.path.getsize(Path(tmp_dir) / f)
                    for f in os.listdir(Path(tmp_dir))
                    if os.path.isfile(Path(tmp_dir) / f)
                )

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        path.mkdir(exist_ok=True)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)

        torch.jit.save(self.model, path / self.MODEL_NAME)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        model = torch.jit.load(path / cls.MODEL_NAME)
        metadata = LearnerMetadata.read(path)
        return cls(
            torch_model=model,
            network_parameters=ModelParams(**metadata.network_parameters),
            input_tfms=MultiStageTransformation.from_dict(metadata.input_tfms)
            if metadata.input_tfms is not None
            else None,
            device=metadata.device,
        )

    @classmethod
    def from_torch_model(
        cls,
        model: Union[Module, GraphModule],
        network_parameters: ModelParams,
        device: Device,
        input_tfms: Optional[MultiStageTransformation] = None,
        input_data: List[torch.Tensor] = None,
    ):
        if device is Device.GPU:
            input_data = [t.cuda() for t in input_data]

        if not isinstance(model, torch.fx.GraphModule):
            model.eval()
            try:
                model_scripted = symbolic_trace(model)
                model_scripted = torch.jit.script(model_scripted)
            except Exception:
                try:
                    model_scripted = torch.jit.script(model)
                except Exception:
                    model_scripted = torch.jit.trace(model, tuple(input_data))
        else:
            model_scripted = torch.jit.script(model)

        return cls(
            torch_model=model_scripted,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
            input_data=input_data,
            device=device,
        )



from pathlib import Path
from setuptools import setup, find_packages


REQUIREMENTS = [
    "numpy>=1.20.0, <=1.23.1",
    "py-cpuinfo>=8.0.0",
    "PyYAML>=6.0",
    "psutil>=5.0.0",
    "requests>=2.26.0",
    "tqdm>=4.36.0",
    "packaging>=21.3",
    "loguru>=0.5.3",
]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf8")

setup(
    name="nebullvm",
    version="0.7.3",
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    long_description=long_description,
    include_package_data=True,
    long_description_content_type="text/markdown",
)



import copy
import re
from abc import ABC, abstractmethod
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, Any, Callable

import numpy as np
import yaml

from nebullvm.operations.optimizations.compressors.base import Compressor
from nebullvm.optional_modules.neural_compressor import Pruning
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import DataLoader, Dataset, Module
from nebullvm.tools.data import DataManager


def _get_model_framework(model: Any) -> str:
    if isinstance(model, Module):
        return "torch"
    elif isinstance(model, tf.Module) and model is not None:
        return "tensorflow"
    else:
        return "numpy"


class IntelPruningCompressor(Compressor, ABC):
    def __init__(self, config_file: str = None):
        super().__init__(config_file)
        self._temp_dir = mkdtemp()

    @property
    def config_key(self) -> str:
        return "intel_pruning"

    @staticmethod
    def _get_default_config() -> Dict:
        # see https://github.com/intel/neural-compressor/blob/master/neural_compressor/conf/config.py  # noqa
        # for further details
        config = {
            "train": {
                "optimizer": {
                    "SGD": {"learning_rate": 0.001},
                },
                "criterion": {
                    "CrossEntropyLoss": {
                        "reduction": "mean",
                        "from_logits": False,
                    },
                },
                "epoch": 10,
                "start_epoch": 0,
                "end_epoch": 10,
                "iteration": 30,
                "execution_mode": "eager",  # either eager or graph
                # "hostfile": None,  # str for multinode training support
            },
            "approach": {
                "weight_compression": {
                    "initial_sparsity": 0.0,
                    "target_sparsity": 0.60,
                    "start_epoch": 0,
                    "end_epoch": 8,
                    "pruners": [
                        {
                            "start_epoch": 0,
                            "end_epoch": 8,
                            "prune_type": "basic_magnitude",
                        },
                    ],
                }
            },
        }
        return config

    def _prepare_pruning_config(self, model: Any):
        pruning_config = copy.deepcopy(self._config)
        framework = _get_model_framework(model)
        config = {
            "model": {
                "name": model.__class__.__name__,
                "framework": framework if framework != "torch" else "pytorch",
            },
            "evaluation": {"accuracy": {"metric": {"topk": 1}}},
            "device": "cpu",
            "tuning": {
                "random_seed": 1978,
                "tensorboard": False,
                "workspace": {"path": self._temp_dir},
            },
            "pruning": pruning_config,
        }
        path_file = Path(self._temp_dir) / "temp.yaml"
        with open(path_file, "w") as f:
            yaml.dump(config, f)
        with open(path_file, "r+") as f:
            file_str = f.read()
            file_str = re.sub(
                "pruners:\n      - end_epoch:",
                "pruners:\n      - !Pruner\n        end_epoch:",
                file_str,
            )
            f.seek(0)
            f.write(file_str)
        return path_file

    def execute(
        self,
        model: Any,
        train_input_data: DataManager,
        eval_input_data: DataManager,
        metric_drop_ths: float,
        metric: Callable,
    ):
        config_file_pr = self._prepare_pruning_config(model)
        prune = Pruning(str(config_file_pr))
        prune.model = model
        prune.train_dataloader = self._get_dataloader(train_input_data)
        prune.eval_dataloader = self._get_dataloader(eval_input_data)
        self.compressed_model = prune.fit()

        if self.compressed_model is not None:
            error = self._compute_error(
                model, self.compressed_model, eval_input_data, metric
            )
            if error > metric_drop_ths:
                self.compressed_model = None
            else:
                self.new_metric_ths = metric_drop_ths - error

    @abstractmethod
    def _compute_error(
        self,
        model: Any,
        compressed_model: Any,
        eval_input_data: DataManager,
        metric: Callable,
    ):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _get_dataloader(input_data: DataManager):
        raise NotImplementedError


class INCDataset(Dataset):
    def __init__(self, input_data: DataManager):
        self.data = input_data
        self.batch_size = input_data[0][0][0].shape[0]

    def __len__(self):
        return sum([batch_inputs[0].shape[0] for batch_inputs, _ in self.data])

    def __getitem__(self, idx):
        batch_idx = int(idx / self.batch_size)
        item_idx = idx % self.batch_size
        data = tuple([data[item_idx] for data in self.data[batch_idx][0]])
        return data, self.data[batch_idx][1][item_idx]


class TorchIntelPruningCompressor(IntelPruningCompressor):
    @staticmethod
    def _get_dataloader(input_data: DataManager):
        bs = input_data[0][0][0].shape[0]
        ds = INCDataset(input_data)
        dl = DataLoader(ds, bs)
        return dl

    def _compute_error(
        self,
        model: Module,
        compressed_model: Module,
        eval_input_data: DataManager,
        metric: Callable,
    ):
        if len(eval_input_data) == 0:
            return np.inf
        metric_val = 0
        for inputs, y in eval_input_data:
            pred_model = model(*inputs)
            pred_compressed_model = compressed_model(*inputs)
            metric_val += metric(pred_model, pred_compressed_model, y)
        return metric_val / len(eval_input_data)



import logging

try:
    from openvino.runtime import Core, Model, CompiledModel, InferRequest
    from openvino.tools.pot import DataLoader
    from openvino.tools.pot import IEEngine
    from openvino.tools.pot import load_model, save_model
    from openvino.tools.pot import compress_model_weights
    from openvino.tools.pot import create_pipeline
except ImportError:
    Model = CompiledModel = InferRequest = Core = object
    DataLoader = IEEngine = object
    load_model = save_model = compress_model_weights = create_pipeline = None

# Fix openvino issue with logging
# It adds a second handler to the root logger that cause issues
if len(logging.getLogger().handlers) > 1:
    logging.getLogger().removeHandler(logging.getLogger().handlers[-1])



import cpuinfo
from loguru import logger

from nebullvm.operations.optimizations.compilers.utils import (
    tvm_is_available,
    onnxruntime_is_available,
    tensorrt_is_available,
    openvino_is_available,
    bladedisc_is_available,
    deepsparse_is_available,
    intel_neural_compressor_is_available,
    torch_tensorrt_is_available,
)
from nebullvm.tools.base import Device
from nebullvm.tools.utils import gpu_is_available, check_module_version


def torch_is_available() -> bool:
    try:
        import torch  # noqa F401

        if not torch.cuda.is_available() and gpu_is_available():
            logger.warning(
                "Installed PyTorch does not have cuda support. "
                "Please ensure that torch.cuda.is_available() "
                "returns True by installing the proper version "
                "of PyTorch. "
            )

        if not check_module_version(torch, min_version="1.10.0"):
            logger.warning(
                "torch module version must be >= 1.10.0. "
                "Please update it if you want to use it."
            )
            return False
    except ImportError:
        return False
    else:
        return True


def tensorflow_is_available() -> bool:
    try:
        import tensorflow  # noqa F401

        if not check_module_version(tensorflow, min_version="2.7.0"):
            logger.warning(
                "tensorflow module version must be >= 2.7.0. "
                "Please update it if you want to use it."
            )
            return False
    except ImportError:
        return False
    else:
        return True


def onnx_is_available() -> bool:
    try:
        import onnx  # noqa F401

        if not check_module_version(onnx, min_version="1.10.0"):
            logger.warning(
                "onnx module version must be >= 1.10.0. "
                "Please update it if you want to use it."
            )
            return False
        return True
    except ImportError:
        return False


def _onnxmltools_is_available():
    try:
        import onnxmltools  # noqa F401

        if not check_module_version(onnxmltools, min_version="1.11.0"):
            logger.warning(
                "onnxmltools module version must be >= 1.11.0. "
                "Please update it if you want to use the ONNX API "
                "or the ONNX pipeline for PyTorch and Tensorflow."
            )
            return False
        else:
            return True
    except ImportError:
        return False


def _onnxsim_is_available():
    try:
        import onnxsim  # noqa F401

        return True
    except ImportError:
        return False


def _polygraphy_is_available():
    try:
        import polygraphy.cuda  # noqa F401

        return True
    except ImportError:
        return False


def tf2onnx_is_available():
    try:
        import tf2onnx  # noqa F401

        return True
    except ImportError:
        return False


def check_dependencies(device: Device):
    missing_frameworks = []
    missing_suggested_compilers = []
    missing_optional_compilers = []
    missing_dependencies = []

    processor = cpuinfo.get_cpu_info()["brand_raw"].lower()

    if not onnx_is_available():
        missing_frameworks.append("onnx")

    if not tvm_is_available():
        missing_optional_compilers.append("tvm")
    if not onnxruntime_is_available():
        missing_suggested_compilers.append("onnxruntime")
    elif not _onnxmltools_is_available():
        missing_dependencies.append("onnxmltools")
    if device is Device.GPU:
        if not tensorrt_is_available():
            missing_suggested_compilers.append("tensorrt")
        else:
            if not _onnxsim_is_available():
                missing_dependencies.append("onnxsim")
            elif not _polygraphy_is_available():
                missing_dependencies.append("polygraphy")
    if device is Device.CPU:
        if not openvino_is_available() and "intel" in processor:
            missing_suggested_compilers.append("openvino")

    if torch_is_available():
        if not tvm_is_available():
            if "tvm" not in missing_optional_compilers:
                missing_optional_compilers.append("tvm")
        if not bladedisc_is_available():
            missing_optional_compilers.append("torch_blade")

        if device is Device.CPU:
            if not deepsparse_is_available() and "intel" in processor:
                missing_suggested_compilers.append("deepsparse")
            if (
                not intel_neural_compressor_is_available()
                and "intel" in processor
            ):
                missing_suggested_compilers.append("neural_compressor")
        elif device is Device.GPU:
            if not torch_tensorrt_is_available:
                missing_suggested_compilers.append("torch_tensorrt")
    else:
        missing_frameworks.append("torch")

    if tensorflow_is_available():
        if not tf2onnx_is_available():
            missing_dependencies.append("tf2onnx")
    else:
        missing_frameworks.append("tensorflow")

    missing_frameworks = ", ".join(missing_frameworks)
    if len(missing_frameworks) > 0:
        logger.warning(
            f"Missing Frameworks: {missing_frameworks}.\n "
            f"Please install them "
            "to include them in the optimization pipeline."
        )

    missing_suggested_compilers = ", ".join(missing_suggested_compilers)
    if len(missing_suggested_compilers) > 0:
        logger.warning(
            f"Missing Compilers: {missing_suggested_compilers}.\n "
            f"Please install them "
            "to include them in the optimization pipeline."
        )

    missing_dependencies = ", ".join(missing_dependencies)
    if len(missing_dependencies) > 0:
        logger.warning(
            f"Missing Dependencies: {missing_dependencies}.\n "
            f"Without them, some compilers "
            f"may not work properly."
        )



import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Union

from loguru import logger

from nebullvm.config import ONNX_OPSET_VERSION
from nebullvm.optional_modules.tensorflow import tensorflow as tf, tf2onnx
from nebullvm.optional_modules.onnx import onnx
from nebullvm.tools.base import ModelParams
from nebullvm.tools.huggingface import TensorFlowTransformerWrapper


def convert_tf_to_onnx(
    model: Union[tf.Module, tf.keras.Model],
    model_params: ModelParams,
    output_file_path: Union[str, Path],
):
    """Convert TF models into ONNX.

    Args:
        model (Union[tf.Module, tf.keras.Model]): TF model.
        model_params (ModelParams): Info about model parameters.
        output_file_path (Path): Path where storing the output file.
    """

    try:
        if isinstance(model, tf.keras.Model) or (
            isinstance(model, TensorFlowTransformerWrapper)
            and isinstance(model.core_model, tf.keras.Model)
        ):
            return convert_keras_to_onnx(model, model_params, output_file_path)
        else:
            return convert_tf_saved_model_to_onnx(model, output_file_path)
    except Exception:
        logger.warning(
            "Something went wrong during conversion from tensorflow"
            " to onnx model. ONNX pipeline will be unavailable."
        )
        return None


def convert_tf_saved_model_to_onnx(
    model: tf.Module, output_file_path: Union[str, Path]
):
    """Convert TF models into ONNX.
    Args:
        model (tf.Module): TF model.
        output_file_path (Path): Path where storing the output file.
    """
    with TemporaryDirectory() as temp_dir:
        tf.saved_model.save(model, export_dir=temp_dir)

        try:
            subprocess.check_output(["python3", "--version"])
            python_cmd = "python3"
        except subprocess.CalledProcessError:
            python_cmd = "python"

        onnx_cmd = [
            python_cmd,
            "-m",
            "tf2onnx.convert",
            "--saved-model",
            f"{temp_dir}",
            "--output",
            f"{output_file_path}",
            "--opset",
            f"{ONNX_OPSET_VERSION}",
        ]
        subprocess.run(onnx_cmd)
        onnx.load(output_file_path)

    return output_file_path


def convert_keras_to_onnx(
    model: tf.keras.Model,
    model_params: ModelParams,
    output_file_path: Union[str, Path],
):
    """Convert keras models into ONNX.

    Args:
        model (tf.keras.Model): keras model.
        model_params (ModelParams): Model Parameters as input sizes and
            dynamic axis information.
        output_file_path (Path): Path where storing the output file.
    """
    # get data types for each input
    dtypes = [
        model_params.input_infos[i].dtype.value
        for i in range(len(model_params.input_infos))
    ]
    # get input shapes for each input
    shapes = [
        [model_params.batch_size]
        + [int(x) for x in model_params.input_infos[i].size]
        for i in range(len(model_params.input_infos))
    ]
    # set the dynamic axes for each input
    if isinstance(model, TensorFlowTransformerWrapper):
        names = list(model.inputs_types.keys())
    else:
        names = [f"input_{i}" for i in range(len(model_params.input_infos))]

    input_signature = tuple(
        tf.TensorSpec(
            (
                None
                if model_params.dynamic_info is not None
                and dim in model_params.dynamic_info.inputs[i]
                else shape[dim]
                for dim in range(len(shape))
            ),
            dtype,
            name=name,
        )
        for i, (shape, dtype, name) in enumerate(zip(shapes, dtypes, names))
    )

    onnx_model, _ = tf2onnx.convert.from_keras(
        model,
        input_signature,
        opset=ONNX_OPSET_VERSION,
        output_path=output_file_path,
    )

    return output_file_path



try:
    import absl.logging

    absl.logging.set_verbosity(absl.logging.ERROR)
except Exception:
    pass


class Model:
    pass


class Dataset:
    pass


class Keras:
    Model = Model


class Module:
    pass


class Tensor:
    pass


class data:
    Dataset = Dataset


class Tensorflow:
    Module = Module
    Tensor = Tensor
    keras = Keras()
    data = data

    @staticmethod
    def function(**kwargs):
        return lambda x: x


try:
    import tensorflow  # noqa F401

    tensorflow.get_logger().setLevel("ERROR")
    tensorflow.autograph.set_verbosity(0)
except (ImportError, AttributeError):
    tensorflow = Tensorflow


try:
    import tf2onnx  # noqa F401

    tf2onnx.logging.set_level("ERROR")
    tf2onnx.logging.set_tf_verbosity("ERROR")
except ImportError:
    tf2onnx = object



from typing import Union, List, Tuple, Any, Optional, Dict

import numpy as np

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.tools.base import InputInfo, DataType, Device


def get_outputs_sizes_tf(
    tf_model: Union[tf.Module, tf.keras.Model],
    input_tensors: List[tf.Tensor],
    device: Device,
) -> List[Tuple[int, ...]]:
    with tf.device(device.value):
        outputs = tf_model(input_tensors)
    if isinstance(outputs, tf.Tensor) and outputs is not None:
        return [tuple(outputs.shape)]
    return [tuple(x.shape) for x in outputs]


def create_model_inputs_tf(
    batch_size: int, input_infos: List[InputInfo]
) -> List[tf.Tensor]:
    return [
        tf.random_normal_initializer()(
            shape=(batch_size, *input_info.size[1:], input_info.size[0])
        )
        if input_info.dtype is DataType.FLOAT32
        else tf.random.uniform(
            shape=(batch_size, *input_info.size[1:], input_info.size[0]),
            minval=input_info.min_value or 0,
            maxval=input_info.max_value or 100,
            dtype=tf.int32,
        )
        for input_info in input_infos
    ]


def run_tf_model(
    model: tf.Module,
    input_tensors: Tuple[tf.Tensor],
    device: Device,
) -> Tuple[tf.Tensor]:
    with tf.device(device.value):
        pred = model(input_tensors)
    if isinstance(pred, tf.Tensor):
        pred = (pred,)
    return pred


def _extract_dynamic_axis(
    tf_model: tf.Module,
    dataset: List[Tuple[Tuple[tf.Tensor, ...], Any]],
    input_sizes: List[Tuple[int, ...]],
    batch_size: int,
    device: Device,
    max_data: int = 100,
) -> Optional[Dict]:
    from nebullvm.tools.utils import inspect_dynamic_size

    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, input_data in enumerate(dataset):
        input_tensors = input_data[0]
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, batch_size, dynamic_axis["inputs"]
        )
        outputs = tuple(run_tf_model(tf_model, input_tensors, device))
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape[1:]) for output in outputs]
        inspect_dynamic_size(
            outputs, output_sizes, batch_size, dynamic_axis["outputs"]
        )
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def extract_info_from_tf_data(
    tf_model: tf.Module,
    dataset: List[Tuple[Tuple[tf.Tensor, ...], Any]],
    batch_size: int,
    input_sizes: List[Tuple[int, ...]],
    input_types: List[str],
    dynamic_axis: Dict,
    device: Device,
):
    from nebullvm.tools.utils import ifnone

    input_row = dataset[0][0]
    batch_size = ifnone(batch_size, int(input_row[0].shape[0]))
    input_sizes = ifnone(input_sizes, [tuple(x.shape[1:]) for x in input_row])
    input_types = ifnone(
        input_types,
        [
            "int32"
            if x.dtype in [tf.int32, np.int32]
            else "int64"
            if x.dtype in [tf.int64, np.int64]
            else "float32"
            for x in input_row
        ],
    )
    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(
            tf_model, dataset, input_sizes, batch_size, device
        ),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def tensorflow_is_gpu_available():
    return len(tf.config.list_physical_devices("GPU")) > 0


def tensorflow_get_gpu_name():
    gpu_devices = tf.config.list_physical_devices("GPU")
    if gpu_devices:
        details = tf.config.experimental.get_device_details(gpu_devices[0])
        details.get("device_name", "Unknown GPU")
        return details["device_name"]
    else:
        return "Unknown GPU"



from typing import List, Tuple

import numpy as np

from nebullvm.optional_modules.tensor_rt import (
    tensorrt as trt,
    IInt8EntropyCalibrator2,
    polygraphy,
)
from nebullvm.tools.base import QuantizationType, ModelParams
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    HalfPrecisionTransformation,
)


def quantize_tensorrt(
    quantization_type: QuantizationType,
    model_params: ModelParams,
    config,
    input_tfms: MultiStageTransformation,
    input_data: List[Tuple[np.ndarray, ...]] = None,
):
    if quantization_type is QuantizationType.HALF:
        config.set_flag(trt.BuilderFlag.FP16)
        input_tfms.append(HalfPrecisionTransformation())
    elif quantization_type is QuantizationType.STATIC:
        assert input_data is not None, (
            "You need to specify the calibration data for "
            "performing static quantization."
        )
        calibrator = TensorRTCalibrator(
            batch_size=model_params.batch_size,
            input_data=input_data,
        )
        config.set_flag(trt.BuilderFlag.INT8)
        config.int8_calibrator = calibrator

    return config


class TensorRTCalibrator(IInt8EntropyCalibrator2):
    def __init__(
        self, batch_size: int, input_data: List[Tuple[np.ndarray, ...]]
    ):
        super(TensorRTCalibrator, self).__init__()
        self._bs = batch_size
        self.batches = (x for x in input_data)

    def get_batch(self, names):
        cuda_stream = polygraphy.Stream()
        try:
            data = next(self.batches)

            cuda_data = []
            for input_tensor in data:
                device_array = polygraphy.DeviceArray()
                device_array.resize(input_tensor.shape)
                device_array.copy_from(
                    host_buffer=input_tensor, stream=cuda_stream
                )
                cuda_data.append(device_array)

            return [input_tensor.ptr for input_tensor in cuda_data]
        except StopIteration:
            return None

    def get_batch_size(self):
        return self._bs

    def read_calibration_cache(self):
        return None

    def write_calibration_cache(self, cache):
        return None



from typing import Sequence, List, Tuple, Any, Union, Iterable

import numpy as np
from loguru import logger

from nebullvm.config import MIN_DIM_INPUT_DATA
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch, Dataset, DataLoader
from nebullvm.tools.onnx import convert_to_numpy


class DataManager:
    """Class for managing the user data in nebullvm.

    Attributes:
        data_reader(Sequence): Object implementing the __getitem__, the
            __len__ and the __iter__/__next__ APIs. It should read the
            user data and return tuples of tensors for feeding the models.
    """

    def __init__(self, data_reader: Sequence):
        self._data_reader = data_reader
        self._pointer = 0
        self.train_idxs = []
        self.test_idxs = []

    def __getitem__(self, item):
        return self._data_reader[item]

    def __len__(self):
        return len(self._data_reader)

    def __iter__(self):
        self._pointer = 0
        return self

    def __next__(self):
        if self._pointer < len(self):
            data = self[self._pointer]
            self._pointer += 1
            return data
        else:
            raise StopIteration

    def get_numpy_list(
        self, n: int = None, shuffle: bool = False, with_ys: bool = False
    ) -> Union[
        List[Tuple[np.ndarray, ...]], Tuple[List[Tuple[np.ndarray, ...]], List]
    ]:
        if n is None:
            n = len(self)
        if not with_ys:
            return [
                tuple(convert_to_numpy(x) for x in tuple_)
                for tuple_ in self.get_list(n, shuffle)
            ]
        else:
            xs, ys = self.get_list(n, shuffle, with_ys=True)
            return [
                tuple(convert_to_numpy(x) for x in tuple_) for tuple_ in xs
            ], ys

    def get_list(
        self, n: int = None, shuffle: bool = False, with_ys: bool = False
    ) -> Union[List[Tuple[Any, ...]], Tuple[List[Tuple[Any, ...]], List]]:
        if n is None:
            n = len(self)
        if shuffle:
            idx = np.random.choice(len(self), n, replace=n > len(self))
        else:
            idx = np.arange(0, min(n, len(self)))
            if n > len(self):
                np.random.seed(0)
                idx = np.concatenate(
                    [
                        idx,
                        np.random.choice(
                            len(self), n - len(self), replace=True
                        ),
                    ]
                )
        if not with_ys:
            return [self[i][0] for i in idx]

        ys, xs = [], []
        for i in idx:
            x, y = self[i] if len(self[i]) > 1 else (self[i][0], None)
            xs.append(x)
            ys.append(y)
        return xs, ys

    @classmethod
    def from_iterable(cls, iterable: Iterable, max_length: int = 500):
        return cls([x for i, x in enumerate(iterable) if i < max_length])

    @classmethod
    def from_dataloader(
        cls,
        dataloader: Union[DataLoader, tf.data.Dataset],
        max_length: int = 500,
    ):
        batch_size = (
            dataloader.batch_size
            if isinstance(dataloader, DataLoader)
            else dataloader._batch_size
        )

        if batch_size > max_length:
            raise ValueError(
                f"Batch size ({dataloader.batch_size}) is greater than "
                f"max_length ({max_length})."
            )
        data_manager = []
        warning_label = False
        for i, batch in enumerate(dataloader):
            if i * batch_size >= max_length:
                break

            if isinstance(batch, (list, tuple)):
                if len(batch) == 1:
                    data_manager.append((batch, None))
                elif len(batch) == 2:
                    if isinstance(batch[0], tuple):
                        data_manager.append((batch[0], batch[1]))
                    elif isinstance(batch[0], (torch.Tensor, tf.Tensor)):
                        warning_label = True
                        data_manager.append(((batch[0],), batch[1]))
                    else:
                        raise ValueError(
                            "The first element of the batch should be a "
                            "tuple or a torch.Tensor"
                        )
                else:
                    warning_label = True
                    data_manager.append(
                        (tuple(t for t in batch[:-1]), batch[-1])
                    )
            elif isinstance(batch, (torch.Tensor, tf.Tensor)):
                data_manager.append(((batch,), None))
            else:
                raise ValueError(
                    "The batch should be a tuple, a list or a Tensor"
                )

        if warning_label:
            logger.warning(
                "The provided dataloader returns a tuple of tensors"
                "for each batch. The last tensor in the tuple will "
                "be considered as the label. "
                "To avoid this warning, the dataloader should return "
                "a tuple for each batch, where the first element is "
                "a tuple containing the inputs and the second element "
                "is a tensor containing the label."
            )

        return cls(data_manager)

    def get_split(self, split_type="train"):
        return (
            DataManager([self[i] for i in self.train_idxs])
            if split_type == "train"
            else DataManager([self[i] for i in self.test_idxs])
        )

    def split(self, split_pct: float, shuffle: bool = False):
        if shuffle:
            idx = np.random.choice(len(self), len(self), replace=False)
        else:
            idx = np.arange(len(self))

        n = int(round(len(idx) * split_pct))

        if len(self) < MIN_DIM_INPUT_DATA:
            logger.warning(
                f"Not enough data for splitting the DataManager. "
                f"You should provide at least {MIN_DIM_INPUT_DATA} "
                f"data samples to allow a good split between train "
                f"and test sets. Compression, calibration and precision "
                f"checks will use the same data."
            )
            self.train_idxs = idx
            self.test_idxs = idx
        else:
            self.train_idxs = idx[:n]
            self.test_idxs = idx[n:]


class PytorchDataset(Dataset):
    def __init__(self, input_data: DataManager, has_labels: bool = False):
        self.data = input_data
        self.has_labels = has_labels
        self.batch_size = input_data[0][0][0].shape[0]

    def __len__(self):
        return sum([batch_inputs[0].shape[0] for batch_inputs, _ in self.data])

    def __getitem__(self, idx):
        batch_idx = int(idx / self.batch_size)
        item_idx = idx % self.batch_size
        data = tuple([data[item_idx] for data in self.data[batch_idx][0]])

        if self.has_labels:
            return data, self.data[batch_idx][1][item_idx]
        else:
            return data



import subprocess
from pathlib import Path
from typing import Tuple, List, Union

import numpy as np

from nebullvm.config import QUANTIZATION_DATA_NUM
from nebullvm.operations.optimizations.compilers.base import Compiler
from nebullvm.operations.optimizations.compilers.quantizations.openvino import (  # noqa: E501
    quantize_openvino,
)
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.optional_modules.openvino import (
    Core,
    CompiledModel,
)
from nebullvm.tools.base import (
    QuantizationType,
    ModelParams,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import get_input_names
from nebullvm.tools.transformations import MultiStageTransformation


class OpenVINOCompiler(Compiler):
    supported_ops = {
        "cpu": [
            None,
            QuantizationType.STATIC,
            QuantizationType.HALF,
        ],
        "gpu": [],
    }

    def __init__(self):
        super().__init__()

    def execute(
        self,
        model: Union[str, Path],
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Compile the input model using OpenVINO library.

        Args:
            model (str): The onnx model path.
            model_params (ModelParams): The model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)
        train_input_data = input_data.get_split("train").get_numpy_list(
            QUANTIZATION_DATA_NUM
        )

        cmd = [
            "mo",
            "--input_model",
            str(model),
            "--output_dir",
            str(Path(model).parent),
            "--input",
            ",".join(get_input_names(model)),
            "--input_shape",
            ",".join(
                [
                    f"{list((model_params.batch_size,) + shape)}"
                    for shape in model_params.input_sizes
                ]
            ),
        ]

        if quantization_type is QuantizationType.DYNAMIC:
            return None

        if quantization_type is QuantizationType.HALF:
            cmd = cmd + ["--data_type", "FP16"]

        process = subprocess.Popen(cmd)
        process.wait()
        base_path = Path(model).parent
        openvino_model_path = base_path / f"{Path(model).stem}.xml"
        openvino_model_weights = base_path / f"{Path(model).stem}.bin"

        if quantization_type not in [QuantizationType.HALF, None]:
            openvino_model_path, openvino_model_weights = self._quantize_model(
                model_topology=str(openvino_model_path),
                model_weights=str(openvino_model_weights),
                input_names=get_input_names(model),
                input_data=train_input_data,
            )

        self.compiled_model = str(
            Path(openvino_model_path).parent / Path(openvino_model_path).stem
        )

    def _compile_model(
        self,
        model_name: str,
        model_weights: str,
        network_parameters: ModelParams,
    ) -> CompiledModel:
        core = Core()
        model = core.read_model(model=model_name, weights=model_weights)

        dynamic_shape = self._get_dynamic_shape(model, network_parameters)

        if dynamic_shape is not None:
            model.reshape(dynamic_shape)

        return core.compile_model(model=model, device_name="CPU")

    @staticmethod
    def _quantize_model(
        model_topology: str,
        model_weights: str,
        input_data: List[Tuple[np.ndarray, ...]],
        input_names: List[str],
    ) -> Tuple[str, str]:
        return quantize_openvino(
            model_topology, model_weights, input_data, input_names
        )



from typing import (
    List,
    Dict,
    Sequence,
    Optional,
)

import numpy as np

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.optional_modules.huggingface import (
    PreTrainedTokenizer,
    PreTrainedModel,
)
from nebullvm.tools.base import Device
from nebullvm.tools.huggingface import (
    get_output_structure_from_dict,
    get_output_structure_from_text,
    PyTorchTransformerWrapper,
    TensorFlowTransformerWrapper,
)
from nebullvm.tools.utils import is_dict_type


class _HFTextDataset(Sequence):
    def __init__(
        self,
        input_texts: List,
        ys: Optional[List],
        keywords: List[str],
        batch_size: int,
        tokenizer: PreTrainedTokenizer,
        tokenizer_args: Dict,
    ):
        self._input_texts = input_texts
        self._ys = ys
        self._bs = batch_size
        self._keys = keywords
        self._tokenizer = tokenizer
        if self._tokenizer.pad_token is None:
            self._tokenizer.pad_token = self._tokenizer.eos_token
        _tokenizer_args = {"truncation": True, "padding": True}
        _tokenizer_args.update(tokenizer_args)
        self._tokenizer_args = _tokenizer_args

    def __getitem__(self, item: int):
        pointer = self._bs * item
        if pointer >= len(self._input_texts):
            raise IndexError
        mini_batch = self._input_texts[
            pointer : pointer + self._bs  # noqa E203
        ]
        if self._ys is not None:
            mini_batch_y = self._ys[pointer : pointer + self._bs]  # noqa E203
        else:
            mini_batch_y = None
        encoded_inputs = self._tokenizer(mini_batch, **self._tokenizer_args)
        return tuple(encoded_inputs[key] for key in self._keys), mini_batch_y

    def __len__(self):
        return len(self._input_texts) // self._bs


class _HFDictDataset(Sequence):
    def __init__(
        self,
        input_data: List,
        ys: Optional[List],
        keywords: List[str],
    ):
        self._input_data = input_data
        self._ys = ys
        self._keys = keywords

    def __getitem__(self, item: int):
        pointer = item
        if pointer >= len(self._input_data):
            raise IndexError
        mini_batch = self._input_data[pointer]
        if self._ys is not None:
            mini_batch_y = self._ys[pointer]
        else:
            mini_batch_y = None
        return (
            tuple(self._concatenate(mini_batch, key) for key in self._keys),
            mini_batch_y,
        )

    def __len__(self):
        return len(self._input_data)

    @staticmethod
    def _concatenate(mini_batch, key):
        if isinstance(mini_batch[key], torch.Tensor):
            return torch.concat([mini_batch[key]])
        elif isinstance(mini_batch[key], tf.Tensor):
            return tf.concat([mini_batch[key]], 0)
        else:
            return np.concatenate([mini_batch[key]])


def convert_hf_model(
    model: PreTrainedModel,
    input_data: List,
    device: Device,
    tokenizer: Optional[PreTrainedTokenizer] = None,
    tokenizer_args: Optional[Dict] = None,
    batch_size: int = 1,
    **kwargs,
):
    if is_dict_type(input_data[0]):
        # already tokenized data
        if "labels" in input_data[0]:
            labels = [data.pop("labels") for data in input_data]
        else:
            labels = None
        input_example = input_data[0]
        output_structure, output_type = get_output_structure_from_dict(
            input_example=input_example,
            model=model,
            device=device,
        )
        input_data = _HFDictDataset(
            input_data=input_data,
            ys=labels,
            keywords=list(input_example.keys()),
        )

    else:
        assert tokenizer is not None, (
            "Tokenizer is needed when passing data in string format. Please "
            "provide the tokenizer as keyword argument."
        )
        if tokenizer_args is None:
            tokenizer_args = {}
        if not isinstance(input_data[0], str):
            ys = [data[1] for data in input_data]
            input_data = [data[0] for data in input_data]
        else:
            ys = None
        output_structure, output_type = get_output_structure_from_text(
            text=input_data[0],
            model=model,
            tokenizer=tokenizer,
            tokenizer_args=tokenizer_args,
            device=device,
        )
        input_example = tokenizer(input_data, **tokenizer_args)
        input_data = _HFTextDataset(
            input_texts=input_data,
            ys=ys,
            keywords=list(input_example.keys()),
            batch_size=batch_size,
            tokenizer=tokenizer,
            tokenizer_args=tokenizer_args,
        )
    if isinstance(model, torch.nn.Module):
        wrapper_model = PyTorchTransformerWrapper(
            core_model=model, encoded_input=input_example
        )
    else:
        wrapper_model = TensorFlowTransformerWrapper(
            core_model=model, encoded_input=input_example
        )

    return (
        wrapper_model,
        input_data,
        list(wrapper_model.inputs_types.keys()),
        output_structure,
        output_type,
    )



import os
from pathlib import Path
from typing import Tuple, Optional, Any, Callable

import torch
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import Model, layers
from transformers import AlbertModel, AlbertTokenizer

from nebullvm.config import TRAIN_TEST_SPLIT_RATIO, CONSTRAINED_METRIC_DROP_THS
from nebullvm.operations.conversions.huggingface import convert_hf_model
from nebullvm.operations.conversions.pytorch import convert_torch_to_onnx
from nebullvm.operations.measures.measures import (
    LatencyOriginalModelMeasure,
    MetricDropMeasure,
)
from nebullvm.operations.measures.utils import compute_relative_difference
from nebullvm.tools.base import (
    DeepLearningFramework,
    ModelParams,
    Device,
    QuantizationType,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation
from nebullvm.tools.utils import gpu_is_available, extract_info_from_data

INPUT_SHAPE = (3, 256, 256)
OUTPUT_SHAPE = (2,)
STATIC_BATCH_SIZE = 1
DYNAMIC_BATCH_SIZE = 2


class TestModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(
            in_channels=3, out_channels=64, kernel_size=3
        )
        self.relu1 = torch.nn.ReLU()
        self.conv2 = torch.nn.Conv2d(
            in_channels=64, out_channels=32, kernel_size=3
        )
        self.relu2 = torch.nn.ReLU()
        self.fcn = torch.nn.Linear(32, 2)

    def forward(self, input_tensor_0, input_tensor_1):
        x0 = self.relu2(self.conv2(self.relu1(self.conv1(input_tensor_0))))
        x1 = self.relu2(self.conv2(self.relu1(self.conv1(input_tensor_1))))
        x = x0 + x1
        x = self.fcn(x.mean(dim=(-2, -1)).view(-1, 32))
        return x


def tensorflow_model():
    input_0 = keras.Input(shape=(256, 256, 3))
    input_1 = keras.Input(shape=(256, 256, 3))
    x0 = layers.Conv2D(64, kernel_size=(3, 3), activation="relu")(input_0)
    x1 = layers.Conv2D(64, kernel_size=(3, 3), activation="relu")(input_1)
    x0 = layers.Conv2D(32, kernel_size=(3, 3), activation="relu")(x0)
    x1 = layers.Conv2D(32, kernel_size=(3, 3), activation="relu")(x1)
    x = x0 + x1
    y = layers.Dense(2, activation="softmax")(x)
    return Model(inputs=[input_0, input_1], outputs=y)


def _build_static_model(
    framework: DeepLearningFramework = DeepLearningFramework.PYTORCH,
) -> Tuple[torch.nn.Module, ModelParams]:
    model_params = {
        "batch_size": STATIC_BATCH_SIZE,
        "input_infos": [
            {"size": INPUT_SHAPE, "dtype": "float32"},
            {"size": INPUT_SHAPE, "dtype": "float32"},
        ],
        "output_sizes": [OUTPUT_SHAPE],
    }
    model_params = ModelParams(**model_params)
    if framework == DeepLearningFramework.PYTORCH:
        model = TestModel()
    elif framework == DeepLearningFramework.TENSORFLOW:
        model = tensorflow_model()
    else:
        raise NotImplementedError
    return model, model_params


def _build_dynamic_model(
    framework: DeepLearningFramework,
) -> Tuple[torch.nn.Module, ModelParams]:
    model_params = {
        "batch_size": DYNAMIC_BATCH_SIZE,
        "input_infos": [
            {"size": INPUT_SHAPE, "dtype": "float32"},
            {"size": INPUT_SHAPE, "dtype": "float32"},
        ],
        "output_sizes": [OUTPUT_SHAPE],
        "dynamic_info": {
            "inputs": [{0: "batch_size"}, {0: "batch_size"}],
            "outputs": [{0: "batch_size"}],
        },
    }
    if framework == DeepLearningFramework.PYTORCH:
        model = TestModel()
    elif framework == DeepLearningFramework.TENSORFLOW:
        model = tensorflow_model()
    else:
        raise NotImplementedError()
    return model, ModelParams(**model_params)


def get_torch_model(dynamic: bool = False):
    if dynamic:
        model, model_params = _build_dynamic_model(
            DeepLearningFramework.PYTORCH
        )
    else:
        model, model_params = _build_static_model(
            DeepLearningFramework.PYTORCH
        )
    return model, model_params


def get_tensorflow_model(dynamic: bool = False):
    if dynamic:
        model, model_params = _build_dynamic_model(
            DeepLearningFramework.TENSORFLOW
        )
    else:
        model, model_params = _build_static_model(
            DeepLearningFramework.TENSORFLOW
        )
    return model, model_params


def get_huggingface_model(temp_dir: str, dl_framework: DeepLearningFramework):
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = AlbertModel.from_pretrained("albert-base-v1")

    text = "Short text you wish to process"
    encoded_input = tokenizer(text, return_tensors="pt")
    device = Device.GPU if gpu_is_available() else Device.CPU

    (
        model,
        input_data,
        input_names,
        output_structure,
        output_type,
    ) = convert_hf_model(model, [encoded_input], device=device)

    input_data = DataManager(input_data)
    input_data.split(TRAIN_TEST_SPLIT_RATIO)

    # Benchmark original model
    benchmark_orig_model_op = LatencyOriginalModelMeasure()
    benchmark_orig_model_op.to(device).execute(
        model=model,
        input_data=input_data.get_split("test"),
        dl_framework=dl_framework,
    )

    model_outputs = benchmark_orig_model_op.get_result()[0]

    model_path = os.path.join(temp_dir, "test_model.onnx")

    model_params = extract_info_from_data(
        model, input_data, dl_framework, None, device
    )

    device = Device.GPU if gpu_is_available() else Device.CPU
    convert_torch_to_onnx(
        model, input_data, model_params, Path(model_path), device
    )

    return (
        model_path,
        model_params,
        output_structure,
        input_names,
        output_type,
        input_data,
        model_outputs,
    )


def initialize_model(
    dynamic: bool,
    metric: Optional[str],
    output_library: DeepLearningFramework,
    device: Device,
):
    torch_device = torch.device("cuda" if device is Device.GPU else "cpu")
    batch_size = DYNAMIC_BATCH_SIZE if dynamic else STATIC_BATCH_SIZE

    if output_library == DeepLearningFramework.PYTORCH:
        model, model_params = get_torch_model(dynamic)

        input_data = DataManager(
            [
                (
                    (
                        torch.randn(batch_size, *INPUT_SHAPE).to(torch_device),
                        torch.randn(batch_size, *INPUT_SHAPE).to(torch_device),
                    ),
                    torch.zeros(batch_size, dtype=torch.long),
                )
            ]
        )
    elif output_library == DeepLearningFramework.TENSORFLOW:
        model, model_params = get_tensorflow_model(dynamic)
        input_data = DataManager(
            [
                (
                    (
                        tf.random_normal_initializer()(
                            shape=(
                                batch_size,
                                *INPUT_SHAPE[1:],
                                INPUT_SHAPE[0],
                            )
                        ),
                        tf.random_normal_initializer()(
                            shape=(
                                batch_size,
                                *INPUT_SHAPE[1:],
                                INPUT_SHAPE[0],
                            )
                        ),
                    ),
                    [0 for _ in range(batch_size)],
                )
            ]
        )

    input_data.split(TRAIN_TEST_SPLIT_RATIO)
    input_tfms = MultiStageTransformation([])

    # Benchmark original model
    benchmark_orig_model_op = LatencyOriginalModelMeasure()
    benchmark_orig_model_op.to(device).execute(
        model=model,
        input_data=input_data.get_split("test"),
        dl_framework=output_library,
    )

    model_outputs = benchmark_orig_model_op.get_result()[0]

    if metric is not None:
        metric = compute_relative_difference

    return model, input_data, model_params, input_tfms, model_outputs, metric


def check_model_validity(
    optimized_model: Any,
    input_data: DataManager,
    model_outputs: Any,
    metric_drop_ths: float,
    quantization_type: QuantizationType,
    metric: Callable,
) -> bool:
    test_input_data, ys = input_data.get_split("test").get_list(with_ys=True)
    validity_check_op = MetricDropMeasure()
    validity_check_op.execute(
        optimized_model,
        test_input_data,
        model_outputs,
        metric_drop_ths
        if metric_drop_ths is not None
        else CONSTRAINED_METRIC_DROP_THS,
        metric_func=metric
        if quantization_type is not None
        else compute_relative_difference,
        ys=ys,
    )

    print(validity_check_op.get_result()[1])

    return validity_check_op.get_result()[0]
